static const char rcsid[] = "$Id: bulletins.c,v 1.3 2000/09/16 22:06:01 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "includes.h"
#include "log.h"
#include "response.h"
#include "maildrop.h"
#include "utils.h"

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#include <sys/stat.h>
#include <fcntl.h>

#ifdef HAVE_UTIME
# include <utime.h>
#endif

#include <errno.h>

#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) (strlen((dirent)->d_name))
#else
# define dirent direct
# define NAMLEN(dirent) ((dirent)->d_namlen)
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif


extern struct str_maildrop *maildrop;

int addbulletins = 1;
char bulldir[PATH_MAX];
char userbullfile[PATH_MAX];

void add_bulletins(char *homedir) {
	char bullmsg[PATH_MAX];
	struct stat stbuf;
	int fd;
	DIR *dirstream;
	struct dirent *dentry;
	char *tmp;
	time_t modtime, actime;
#ifdef HAVE_UTIME
	struct utimbuf newmtime;
#else
	struct timeval newmtime[2];
#endif
	
	if (!addbulletins) {
		maildrop->md_end_of_adding();
		return;
	}
	if ((strlen(bulldir) + 1) >= sizeof(bullmsg)) {
		pop_log(pop_priority, "unexpected error - contact author "
			"of program");
		send_error("fatal error");
		exit(1);
	}
	strcpy(bullmsg, bulldir);
	strcat(bullmsg, "/");
	tmp = bullmsg + strlen(bullmsg);
	if (expand_dir(userbullfile, homedir, sizeof(userbullfile)) < 0) {
		pop_log(pop_priority, "error while expanding user bulletin "
			"file");
		send_error("fatal error");
		exit(1);
	}
	modtime = 1;
	fd = open(userbullfile, O_RDONLY);
	if ((fd < 0) && (errno == ENOENT)) {
		fd = open(userbullfile, O_RDONLY | O_CREAT | O_EXCL, 0600);
		modtime = 0;
	}
	if (fd < 0) {
	/* It runs with user privileges, so users can do what they want */
		pop_log(pop_priority, "can't open or create file: %.1024s",
			userbullfile);
		pop_error("open");
		send_error("fatal error");
		exit(1);
	}
	if (fstat(fd, &stbuf) < 0) {
		close(fd);
		pop_log(pop_priority, "can't stat file: %.1024s",
			userbullfile);
		pop_error("stat");
		send_error("fatal error");
		exit(1);
	}
	if (modtime)
		modtime = stbuf.st_mtime;
	actime = stbuf.st_atime;
	close(fd); /* We have modification time of user bulletin file */
	if (!S_ISREG(stbuf.st_mode) || (stbuf.st_mode & 2) ||
		(getuid() != stbuf.st_uid)) {
		pop_log(pop_priority, "user bulletin file has wrong mode");
		send_error("fatal error");
		exit(1);
	}
	if ((dirstream = opendir(bulldir)) == NULL) {
		pop_log(pop_priority, "opendir failed: %.1024s", bulldir);
		pop_error("opendir");
		send_error("fatal error");
		exit(1);
	}
	while ((dentry = readdir(dirstream)) != NULL) {
		*tmp = 0;
		if ((NAMLEN(dentry) + strlen(bullmsg) + 1) > sizeof(bullmsg)) {
			closedir(dirstream);
			pop_log(pop_priority, "file name too long");			
			send_error("fatal error");
			exit(1);
		}
		strcpy(tmp, dentry->d_name);
		if (stat(bullmsg, &stbuf) < 0) {
			closedir(dirstream);
			pop_log(pop_priority, "stat failed: %.1024s", bullmsg);
			pop_error("stat");
			send_error("fatal error");
			exit(1);
		}
		if (!S_ISREG(stbuf.st_mode))
			continue;
		if ((fd = open(bullmsg, O_RDONLY)) < 0) {
			closedir(dirstream);
			pop_log(pop_priority, "can't open file: %.1024s",
				bullmsg);
			pop_error("open");
			send_error("fatal error");
			exit(1);
		}
		if (modtime > stbuf.st_mtime) {
			close(fd);
			continue;
		}
		if (maildrop->md_add_message(fd) == 0) {
			close(fd);
			break;
		}
		close(fd);
	}
	closedir(dirstream);
	maildrop->md_end_of_adding();
#ifdef HAVE_UTIME
	newmtime.actime = actime;
	newmtime.modtime = time(NULL);
	if (utime(userbullfile, &newmtime) < 0) {
		pop_log(pop_priority, "can't set modification time of the "
			"file: %.1024s", userbullfile);
		pop_error("utime");
#else
	newmtime[0].tv_usec = newmtime[1].tv_usec = 0;
	newmtime[0].tv_sec = actime;
	newmtime[1].tv_sec = time(NULL);
	if (utimes(userbullfile, newmtime) < 0) {
		pop_error("utimes");
#endif		
		send_error("fatal error");
		exit(1);
	}
}
