static const char rcsid[] = "$Id: cmds.c,v 1.8 2000/08/27 17:23:21 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <signal.h>

#include "includes.h"
#include "cmds.h"
#include "response.h"
#include "log.h"
#include "sktbuff.h"

extern unsigned int autologout_time;


extern void check_wccount(void);

char upcase(char what) { /* toupper() uses locale so is slower */
	if ((what <= 'z') && (what >= 'a'))
		return what - ('a' - 'A');
	else
		return what;
}

const struct s_cmd * cmd_lookup(const struct s_cmd *cmdtbl, char *cmd) {
	int actual = 0, length, tmp;

	while (cmdtbl[actual].name != NULL) {
		length = strlen(cmdtbl[actual].name);
		if (strlen(cmd) < length) {
			actual++;
			continue;
		}
		tmp = 0;
		while ((cmdtbl[actual].name[tmp] == upcase(cmd[tmp])) &&
			(tmp < length))
			tmp++; /* first "length" bytes are the same */
		if ((tmp == length) && ((strlen(cmd) == length) ||
			(cmd[length] == ' ')))
			return &cmdtbl[actual];
		actual++;
	}
	return NULL;
}

void ignore_cmd(char *arg) {
	send_ok("command ignored");
}

void read_command(char *cmd) {
	ssize_t cread;
	char buf[MAXCMDLN + 1];
	
	alarm(autologout_time);
	buf[MAXCMDLN] = 0;
	while (1) {
		if ((cread = skt_fgets(buf, MAXCMDLN)) < 2) {
			if (cread < 0)
				pop_error("read");
			memset(buf, 0, sizeof(buf));
			exit(1);
		}
		if ((buf[cread - 2] != '\r') || (buf[cread - 1] != '\n')) {
			if (cread == MAXCMDLN) {/* command is too long */
				while ((buf[cread - 2] != '\r') ||
					(buf[cread - 1] != '\n'))
					if ((cread = skt_fgets(buf, MAXCMDLN))
						< 2) {
						if (cread < 0);
							pop_error("read");
						memset(buf, 0, sizeof(buf));
						exit(1);
					}
				memset(buf, 0, sizeof(buf));
				send_error("command too long");
				if (buf_flush() < 0)
					exit(1);
				check_wccount();
				continue;
			} else {
				memset(buf, 0, sizeof(buf));
				exit(1);
			}
		}
		buf[cread - 2] = 0;
		if (strlen(buf) != (cread - 2)) {
			memset(buf, 0, sizeof(buf));
			send_error("NULL character in command");
			if (buf_flush() < 0)
				exit(1);
			check_wccount();
			continue;
		}
		strcpy(cmd, buf); /* strlen(buf) < sizeof(cmd)
		sizeof(buf) = sizeof(cmd) + 1 and \r\n in buf are discarded */
		memset(buf, 0, sizeof(buf));
		break;
	}
	alarm(0);
}
