static const char rcsid[] = "$Id: configoptions.c,v 1.13 2001/04/22 08:57:09 siefca Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include "configfile.h"
#include "utils.h"
#include "log.h"
#include "response.h"

extern void child_corrupted(void);

struct str_option options_set[] = {
#ifdef ALLOWROOTLOGIN
	{"AllowRootLogin", OP_BOOLEAN, &allow_root, 0, NULL},
#endif
#ifdef AUTHONLY
	{"AuthOnly", OP_BOOLEAN, &authonly, 0, NULL},
	{"AuthOnlyUser", OP_STRING, &authonly_user, MAXARGLN + 1, NULL},
	{"AuthOnlyGroup", OP_STRING, &authonly_group, MAXARGLN + 1, NULL},
	{"AuthOnlyFakeHome", OP_STRING, &authonly_fakehome, PATH_MAX, NULL},
#endif
#ifdef USERCONFIG
	{"UserOverride", OP_BOOLEAN, &useroverride, 0, NULL},
#endif
#ifdef APOP
	{"AllowAPOP", OP_BOOLEAN, &allowapop, 0, NULL},	
#endif
	{"MailDropName", OP_STRING, &maildrop_name, PATH_MAX, NULL},
#ifdef CONFIGFILE	
	{"MailDropType", OP_STRING, &maildrop_type, MAXMDTYPENAMELENGTH,
		&check_maildrop_type},
#else
	{"MailDropType", OP_STRING, &maildrop_type, MAXMDTYPENAMELENGTH, NULL},
#endif	
#ifdef APOP
	{"APOPServerName", OP_STRING, &apopservername, 256, NULL},
#endif
	{"AutoLogoutTime", OP_PERIOD, &autologout_time, 0, NULL},
	{"ChangeGID", OP_BOOLEAN, &changegid, 0, NULL},
	{"WrongCommandsLimit", OP_PERIOD, &wccount, 0, NULL},
#ifdef BULLETINS
	{"UserBullFile", OP_STRING, &userbullfile, PATH_MAX, NULL},
	{"BulletinDirectory", OP_STRING, &bulldir, PATH_MAX, NULL},
	{"AddBulletins", OP_BOOLEAN, &addbulletins, 0, NULL},
#endif
#ifdef EXPIRATION
	{"ReadExpire", OP_EXPIRE, &rexp, 0, NULL},
	{"UnreadExpire", OP_EXPIRE, &unrexp, 0, NULL},
#endif
#ifdef MAPPING
	{"UserMapFile", OP_STRING, &sp_mapfile, PATH_MAX, NULL},
	{"DoMapping", OP_BOOLEAN, &domapping, 0, NULL},
	{"RequiredMapping", OP_BOOLEAN, &reqmapping, 0, NULL},
	{"PreAuthMap", OP_STRING, &preauth_map_format, MAXARGLN + 1, NULL},
#endif
#ifdef NONIPVIRTUALS
	{"AllowNonIP", OP_BOOLEAN, &allownonip, 0, NULL},
#endif
#ifdef CREATEMAILDROP
	{"CreateMailDrop", OP_BOOLEAN, &createmaildrop, 0, NULL},
#endif
#ifdef STATISTICS
	{"LogStatistics", OP_BOOLEAN, &logstatistics, 0, NULL},
#endif
#ifdef LOG_EXTEND
	{"AuthLog", OP_STRING, &authlog, PATH_MAX, NULL},
#endif
#ifdef SPSASL
	{"SASLMinSSF", OP_PERIOD, &min_ssf, 0, NULL},
	{"SASLMaxSSF", OP_PERIOD, &max_ssf, 0, NULL},
	{"SASLAllow", OP_BOOLEAN, &allowsasl, 0, NULL},
	{"SASLRealm", OP_STRING, &saslrealm, PATH_MAX, NULL},
#endif
#ifdef SPTLS
	{"TLSCertFile", OP_STRING, &certfile, PATH_MAX, NULL},
	{"TLSKeyFile", OP_STRING, &keyfile, PATH_MAX, NULL},
	{"TLSDCertFile", OP_STRING, &dcertfile, PATH_MAX, NULL},
	{"TLSDKeyFile", OP_STRING, &dkeyfile, PATH_MAX, NULL},
	{"TLSCAFile", OP_STRING, &cafile, PATH_MAX, NULL},
	{"TLSCAPath", OP_STRING, &capath, PATH_MAX, NULL},
	{"TLSDH512ParamFile", OP_STRING, &dh512paramfile, PATH_MAX, NULL},
	{"TLSDH1024ParamFile", OP_STRING, &dh1024paramfile, PATH_MAX, NULL},
	{"TLSCipherList", OP_STRING, &cipherlist, PATH_MAX, NULL},
	{"TLSAskCert", OP_BOOLEAN, &askcert, 0, NULL},
	{"TLSRequireCert", OP_BOOLEAN, &requirecert, 0, NULL},
	{"TLSVerifyDepth", OP_PERIOD, &verifydepth, 0, NULL},
	{"TLSAllow", OP_BOOLEAN, &allowtls, 0, NULL},
	{"TLSRequire", OP_BOOLEAN, &requiretls, 0, NULL},
	{"TLSLogLevel", OP_PERIOD, &tls_loglevel, 0, NULL},
	{"TLSWrap", OP_BOOLEAN, &enabletls, 0, NULL},
	{"TLSRandSource", OP_STRING, &rand_source, PATH_MAX, NULL},
	{"TLSRandBytes", OP_PERIOD, &rand_bytes, 0, NULL},
	{"TLSAllowTmpRSA", OP_BOOLEAN, &allowtmprsa, 0, NULL},
#endif
#ifdef DPUID
	{"DpUIDUser", OP_STRING, &dpuid_user, MAXARGLN + 1, NULL},
	{"DpUIDDBFile", OP_STRING, &dpuid_db, PATH_MAX, NULL},
	{"DpUIDEnable", OP_BOOLEAN, &dpuid_enabled, 0, NULL},
#endif
#ifdef USERPASSWD
	{"RequireUserPassword", OP_BOOLEAN, &requserpasswd, 0, NULL},
	{"AllowUserPassword", OP_BOOLEAN, &allowuserpasswd, 0, NULL},
#endif
	{"LogPriority", OP_STRING, &logpriority, 64, check_logpriority},
	{"AllowUser", OP_BOOLEAN, &allowuser, 0, NULL},
	{"Debug", OP_BOOLEAN, &pop_debug, 0, NULL},
	{NULL, 0, NULL, 0, NULL}
};

void write_configuration(int fd) {
	int tmp = 0;

	while (options_set[tmp].name != NULL) {
	/* This loop sends configuration */
		switch (options_set[tmp].op_type) {
			case OP_STRING:
				if (write_string(fd,
					(char *)options_set[tmp].value) < 0) {
					close(fd);
					pop_error("write");
					send_error("fatal error");
					exit(1);
				}
				break;
			case OP_BOOLEAN:
				if (write_loop(fd, options_set[tmp].value,
					sizeof(int)) != sizeof(int)) {
					close(fd);
					pop_error("write");
					send_error("fatal error");
					exit(1);
				}
				break;
			case OP_PERIOD:
				if (write_loop(fd, options_set[tmp].value,
					sizeof(unsigned int)) !=
					sizeof(unsigned int)) {
					close(fd);
					pop_error("write");
					send_error("fatal error");
					exit(1);
				}
				break;
#ifdef EXPIRATION
			case OP_EXPIRE:
				if (write_loop(fd, options_set[tmp].value,
					sizeof(struct expiration)) != 
					sizeof(struct expiration)) {
					close(fd);
					pop_error("write");
					send_error("fatal error");
					exit(1);
				}
#endif
		}
		tmp++;
	}
}

/* This function works with root privileges */
void read_configuration(int fd) {
	int tmp = 0, tmp2;

	while (options_set[tmp].name != NULL) {
		switch (options_set[tmp].op_type) {
			case OP_STRING:
				if (read_string(fd,
					(char *)options_set[tmp].value,
					options_set[tmp].valuesize) < 0)
					child_corrupted();
				break;
			case OP_BOOLEAN:
				if (read_loop(fd, options_set[tmp].value,
					sizeof(int)) != sizeof(int))
					child_corrupted();
				tmp2 = *((int *)(options_set[tmp].value));
				if ((tmp2 != 0) && (tmp2 != 1))
					child_corrupted();						
				break;
			case OP_PERIOD:
				if (read_loop(fd, options_set[tmp].value,
					sizeof(unsigned int)) !=
					sizeof(unsigned int))
					child_corrupted();
				break;
#ifdef EXPIRATION
			case OP_EXPIRE:
				if (read_loop(fd, options_set[tmp].value,
					sizeof(struct expiration)) !=
					sizeof(struct expiration))
					child_corrupted();
				break;
#endif
		}
		tmp++;
	}
}
