static const char rcsid[] = "$Id: fdfgets.c,v 1.7 2000/09/29 10:31:53 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"

#include "fdfgets.h"

char fgets_buf[10240], *fgets_act;
size_t fgets_count;

char *findcrlf(char *where, size_t length) {
/* strstr stops on NUL character */
	size_t tmp;

	if (length == 0)
		return NULL;
	for (tmp = 0; tmp < (length - 1); tmp++)
		if ((where[tmp] == '\r') && (where[tmp + 1] == '\n'))
			return &where[tmp];
	return NULL;
}

ssize_t gen_fgets(char *s, size_t size, int fd, _read_proc read_proc,
		  char *inbuf, size_t *inbuf_count, char **inbuf_act,
		  size_t inbuf_size, int crlf, size_t *amaxsize) {
	char *le = NULL;
	size_t cread = 0, act;
	size_t maxsize;
	ssize_t result;
	int tmp = (crlf ? 1 : 0);

	while (size > 0) {
		result = 1; /* used to check for EOF */
		if ((*inbuf_count) <= tmp) {
			if ((*inbuf_count) > 0)
				inbuf[0] = **inbuf_act;
			maxsize = inbuf_size - (*inbuf_count);
			if (amaxsize != NULL)
				if (*amaxsize < maxsize)
					maxsize = *amaxsize;
			if (maxsize > 0)
				result = read_proc(fd, inbuf + (*inbuf_count),
					maxsize);
			else
				result = 0;
			if (result < 0)
				return -1;
			if (amaxsize != NULL)
				*amaxsize -= result;
			(*inbuf_act) = inbuf;
			(*inbuf_count) += result;
		}
		if ((*inbuf_count) == 0)
			return cread;
		act = ((size < (*inbuf_count)) ? size : (*inbuf_count));
		if (crlf)
			le = findcrlf((*inbuf_act), act);
		else
			le = memchr((*inbuf_act), '\n', act);
		if (le != NULL)
			act = (le - (*inbuf_act)) + 1 + tmp;
		else {
			if ((act == (*inbuf_count)) && crlf && (result != 0))
				act--;
		}
		memcpy(s, (*inbuf_act), act);
		(*inbuf_act) += act;
		(*inbuf_count) -= act;
		size -= act;
		cread += act;
		s += act;
		if ((le != NULL) || (result == 0))
			return cread;
	}
	return cread;
}

ssize_t fd_fgets(char *s, size_t size, int fd, size_t *maxsize) {
	return gen_fgets(s, size, fd, read, fgets_buf, &fgets_count,
			 &fgets_act, sizeof(fgets_buf), 0, maxsize);
}

void fd_initfgets(void) {
	fgets_count = 0;
	fgets_act = fgets_buf;
}

void fd_cleanup(void) {
	memset(fgets_buf, 0, sizeof(fgets_buf));
}
