static const char rcsid[] = "$Id: hashspool.c,v 1.1 2000/09/22 09:59:18 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"

int qpop_hash(char *what, char *result, char **act, char *where,
	size_t maxlength) {
	int tmp;

	if ((strlen(result) + strlen(*act) + 2) > maxlength)
		return -1;
	/* what is padded with NULs */
	strcat(result, *act);
	result[tmp = strlen(result)] = (((what[0] + what[1] + what[2] +
		what[3]) % 26) + 'a');
	result[tmp + 1] = 0;
	*act = where;
	return 0;
}

/* 
 * nhash code taken from exim Copyright (c) 1999 University of Cambridge
 * Covered by the GNU General Public License
 * original code by Philip Hazel <ph10@cus.cam.ac.uk>
 *
 */

int exim_nhash(char *what, char *result, char **act, char *where,
	size_t maxlength) {

	static int prime[] = {
		2,  3,    5,   7,  11,  13,  17,  19,  23,  29,
		31, 37,   41,  43,  47,  53,  59,  61,  67,  71,
		73, 79,   83,  89,  97, 101, 103, 107, 109, 113
	};
	unsigned long int first, second, total = 0;
	char retval[128];
	char *tmp;
	int i = 0;

	if (!isdigit(*where))
		return -1;
	if ((first = strtoul(where, &tmp, 10)) == ULONG_MAX)
		return -1;
	if (first == 0)
		return -1;
	where = tmp;
	second = 0;
	if (*where == '.')
		if (isdigit(where[1]))
			if ((second = strtoul(where + 1, &tmp, 10)) ==
				ULONG_MAX)
				return -1;
	if (*tmp == '%') /* skip */
		tmp++;
        while (*what != 0) {
		if (i == 0)
			i = sizeof(prime)/sizeof(int) - 1;
		total += prime[i--] * (unsigned int)(*what++);
	}
	if (second == 0) /* If second is unset, just compute one number */
		snprintf(retval, sizeof(retval), "%lu", total % first);
	else { /* Otherwise do a div/mod hash */
		total %= (first * second);
		snprintf(retval, sizeof(retval), "%lu/%lu",
			total/second, total % second);
	}
	if ((strlen(result) + strlen(*act) + strlen(retval) + 1) > maxlength)
		return -1;
	strcat(result, *act);
	strcat(result, retval);
	*act = tmp;
	return 0;
}

/* pjw_hash code taken from ZMailer */

int pjw_hash(char *what, char *result, char **act, char *where,
	size_t maxlength, int single) {
	unsigned long int hash, tmp;
	char retval[8];

	for (hash = 0; *what != 0; what++) {
		hash <<= 4;
		hash += *what;
		if ((tmp = (hash & ~(unsigned long)(0xfffffffL)))) {
			hash ^= tmp >> 24;
			hash ^= tmp;
		}
	}
	if (single)
		snprintf(retval, sizeof(retval), "%c", 'A' + (hash % 26));
	else
		snprintf(retval, sizeof(retval), "%c/%c",
			'A' + ((hash / 26) % 26), 'A' + (hash % 26));
	if ((strlen(result) + strlen(*act) + strlen(retval) + 1) > maxlength)
		return -1;
	strcat(result, *act);
	strcat(result, retval);
	*act = where;
	return 0;
}
