static const char rcsid[] = "$Id: main.c,v 1.36 2001/04/22 08:57:09 siefca Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/resource.h>
#include <pwd.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <signal.h>
#include <ctype.h>

#include "response.h"
#include "const.h"
#include "cmds.h"
#include "maildrop.h"
#ifdef MDMAILBOX
# include "mailbox.h"
#endif
#ifdef MDMAILDIR
# include "maildir.h"
#endif
#include "options.h"
#include "log.h"
#include "userconfig.h"
#ifdef APOP
# include "apop.h"
#endif
#ifdef MAPPING
# include "mapping.h"
#endif
#ifdef SPIPV6
# include "spipv6.h"
#endif
#ifdef BULLETINS
# include "bulletins.h"
#endif

#ifdef SPENCRYPTION
extern int encenabled;
extern int enctrans;
# ifdef SPSASL
#  include "spsasl.h"
extern char sasl_mechanism[];
# endif
# ifdef SPTLS
#  include "sptls.h"
# endif
#endif

#ifdef DPUID
extern char dpuid_log[];
#endif

#include "configfile.h"
#include "authenticate.h"
#include "utils.h"
#include "configoptions.h"
#include "capa.h"
#include "sktbuff.h"
#include "spgetpwnam.h"
#include "fdfgets.h"

#ifdef MD5PASSWORD
extern char sppassword[];
#endif

void get_username(char *);
void end_auth_state(char *);
void do_authentication(char *);
void do_stat(char *);
void do_list(char *);
void do_retrieve(char *);
void do_delete(char *);
void do_reset(char *);
void last_hack(char *);
void do_top(char *);
void do_uidl(char *);
#ifdef APOP
void do_apop(char *);
#endif
void end_trans_state(char *);

const struct s_cmd auth_cmds[] = {
	{"USER", get_username},
	{"PASS", do_authentication},
#ifdef SPSASL
	{"AUTH", do_auth},
#endif
#ifdef SPTLS
	{"STLS", do_stls},
#endif
#ifdef APOP
	{"APOP", do_apop},
#endif
	{"CAPA", print_capa},
	{"QUIT", end_auth_state},
	{NULL, NULL}
};

const struct s_cmd transaction_cmds[] = {
	{"CAPA", print_capa},
	{"STAT", do_stat},
	{"LIST", do_list},
	{"RETR", do_retrieve},
	{"DELE", do_delete},
	{"NOOP", ignore_cmd},
#ifdef LASTCMD
	{"LAST", last_hack},
#endif
	{"RSET", do_reset},
	{"TOP", do_top},
	{"UIDL", do_uidl},
	{"QUIT", end_trans_state},
	{NULL, NULL}
};

const struct s_cmd *cmds[] = {auth_cmds, transaction_cmds};

extern struct str_option options_set[];

int allowuser = 1;
int connection_state;
char username[MAXARGLN + 1];
char password[MAXARGLN + 1];
auth_method_t auth_method;
int gargc;
char **gargv;
char cmd[MAXCMDLN];
extern size_t count;
struct str_maildrop *maildrop;
#ifdef BULLETINS
extern char userbullfile[];
extern char bulldir[];
extern int addbulletins;
#endif
#ifdef EXPIRATION
struct expiration rexp = {0, 0}, unrexp ={0, 0};
#endif
char maildrop_name[PATH_MAX];
char maildrop_type[MAXMDTYPENAMELENGTH];
#ifdef MAPPING
char sp_mapfile[PATH_MAX];
int domapping = 0, reqmapping = 1;
char mapusername[MAXARGLN + 1];
char preauth_map_format[MAXARGLN + 1];
#endif
#ifdef AUTHONLY
struct passwd authonly_pw;
int authonly = 0;
char authonly_fakehome[PATH_MAX];
char authonly_user[MAXARGLN + 1];
char authonly_group[MAXARGLN + 1];
#endif
#ifdef APOP
char apop_secret[MAXARGLN + 1];
char apoptimestamp[MAXRESPLN];
int allowapop = 1;
char apopservername[256] = {0, };
#endif
#ifdef SPSASL
char sasl_secret[SASLSECRETLN];
int sasl_secret_len = 0;
#endif
unsigned int autologout_time = DEFAUTOLOGOUTTIME, wccount = DEFWCCOUNT;
#ifdef ALLOWROOTLOGIN
int allow_root = 0;
#endif
#ifdef USERCONFIG
int useroverride = 1;
#endif
#ifdef NONIPVIRTUALS
int allownonip = 1;
#endif
int auth_finished = 0, auth_successful = -1;
int releasemaildrop = 0;
int changegid = 1;
int tunnel[2] = {-1, -1};
extern int msgnr;
extern struct message *messages;
#ifdef CREATEMAILDROP
int createmaildrop = 0;
#endif
#if defined(LOG_EXTEND) || defined(LOG_CONNECT) || defined(STANDALONE)
char ahname[384];
#endif
#ifdef STATISTICS
int logstatistics = 1;
#endif
#ifdef LOG_EXTEND
char authlog[PATH_MAX];
#endif
#ifdef SPTLS
int enabletls = 0;
#endif
#ifdef SPIPV6
union sp_sockaddr skaddress;
#else
struct sockaddr_in skaddress;
#endif
#ifdef USERPASSWD
int requserpasswd = 0;
int allowuserpasswd = 1;
#endif

void check_wccount(void) {
	if (wccount != 0)
		if ((--wccount) == 0) {
#ifdef LOG_EXTEND
			pop_log(pop_priority, "wrong commands limit exceeded"
				" - %.384s", ahname);
#else
			pop_log(pop_priority, "wrong commands limit exceeded");
#endif
			exit(1);
		}
}

void spop3d_done(void) {
	buf_flush();
	if (tunnel[0] != -1)
		close(tunnel[0]);
	if (tunnel[1] != -1)
		close(tunnel[1]);
#ifdef LOG_EXTEND
	authlog_closelog();
#endif
	memset(password, 0, sizeof(password));
	memset(cmd, 0, sizeof(cmd));
	skt_cleanup(); /* clear socket buffer */
	fd_cleanup();
#ifdef APOP
	memset(apop_secret, 0, sizeof(apop_secret));
#endif
#ifdef MD5PASSWORD
	memset(sppassword, 0, strlen(sppassword));
#endif
	if (releasemaildrop)
		maildrop->md_release();
#ifdef SPENCRYPTION
	close_encpipes();
# ifdef SPSASL
	memset(sasl_secret, 0, sizeof(sasl_secret));
	sasl_close_pipe();
	dispose_sasl();
	done_sasl();
	enc_cleanup();
# endif
# ifdef SPTLS
	tls_stop_servertls();
	tls_done_serverengine();
# endif
#endif
#ifdef HAVE_PAM
	sp_pam_close_session();
#endif
	pop_closelog();
}

void end_trans_state(char *arg)
{
#ifdef STATISTICS
	int deleted = 0, deletedsize = 0, left = 0, leftsize = 0;
#endif
#ifdef EXPIRATION
	int tmp;
	struct expiration *curexp;
	time_t acttime = time(NULL);

	for (tmp = 0; tmp < msgnr; tmp++) {
		if (messages[tmp].cread)
			curexp = &rexp;
		else
			curexp = &unrexp;
		if (curexp->enabled == 0)
			continue;
		if ((messages[tmp].msg_time == -1) ||
		    (messages[tmp].msg_time > acttime)) /* Clock skew??? */
			continue;
		if ((acttime - messages[tmp].msg_time) >= curexp->expperiod)
			messages[tmp].deleted = 1;
	}
#endif /* EXPIRATION */
	maildrop->md_update();

#ifdef STATISTICS
	for (tmp = 0; tmp < msgnr; tmp++)
		if (messages[tmp].deleted) {
			deleted++;
			deletedsize += messages[tmp].crlfsize;
		} else {
			left++;
			leftsize += messages[tmp].crlfsize;
		}
	if (logstatistics)
		pop_log(pop_priority, "Stats: %.40s %d %d %d %d", username, deleted, deletedsize,
			left, leftsize);
#endif
#ifdef MAPPING
	if (domapping)
# ifdef LOG_EXTEND
		pop_log(pop_priority, "session ended for mapped user %.40s - "
			"%.384s", username, ahname);
# else
		pop_log(pop_priority, "session ended for mapped user %.40s",
			username);
# endif
	else
# ifdef LOG_EXTEND
		pop_log(pop_priority, "session ended for user %.40s - %.384s",
			username, ahname);
# else
		pop_log(pop_priority, "session ended for user %.40s",
			username);
# endif
#else /* MAPPING */
# ifdef LOG_EXTEND
	pop_log(pop_priority, "session ended for user %.40s - %.384s",
		username, ahname);
# else
	pop_log(pop_priority, "session ended for user %.40s", username);
# endif
#endif /* MAPPING */

	send_ok("session ended");
	exit(0);
}


void end_auth_state(char *arg) {
#ifdef LOG_EXTEND
	pop_log(pop_priority, "session ended - %.384s", ahname);
#else
	pop_log(pop_priority, "session ended");
#endif
	send_ok("session ended");
	exit(0);
}

void get_username(char *arg) {
	int tmp = 0;

#ifdef SPTLS
	if (requiretls && ((encenabled & 2) == 0)) {
		send_error("you must establish TLS connection first");
		check_wccount();
		return;
	}
#endif
	if (!allowuser) {
		send_error("USER/PASS authentication not allowed");
		check_wccount();
		return;
	}
	if (strlen(arg) >= sizeof(username)) {
		send_error("username too long");
		check_wccount();
	} else 
		if (arg[0] == 0) {
			send_error("missing argument");
			check_wccount();
		} else 
			if (username[0] == 0) {
				strncpy(username, arg, sizeof(username));
			/* Pad with NULs, strlen(arg) < sizeof(username) */
				while (username[tmp] != 0) {
					username[tmp] = tolower(username[tmp]);
					tmp++;
				}
				send_ok("username accepted");
			} else {
				send_error("username already set");
				check_wccount();
			}
}

#ifdef LASTCMD
void last_hack(char *arg) {
	send_ok("0");
}
#endif

void do_stat(char *arg)
{
	md_stat();
}

void do_list(char *arg)
{
	unsigned int number = 1;

	if (arg[0]) {
		if ((sscanf(arg, "%u", &number) != 1) || number == 0) {
			send_error("incorrect argument");
			check_wccount();
		} else
			md_list(number);
	} else
		md_list(0);
}

void do_retrieve(char *arg)
{
	unsigned int number = 1;

	if (!arg[0]) {
		send_error("argument required");
		check_wccount();
		return;
	}
	if ((sscanf(arg, "%u", &number) != 1) || number == 0) {
		send_error("incorrect argument");
		check_wccount();
	} else
		maildrop->md_retrieve(number);
}

void do_delete(char *arg)
{
	unsigned int number = 1;

	if (!arg[0]) {
		send_error("argument required");
		check_wccount();
		return;
	}
	if ((sscanf(arg, "%u", &number) != 1) || number == 0) {
		send_error("incorrect argument");
		check_wccount();
	} else
		md_delete(number);
}

void do_reset(char *arg)
{
	md_reset();
}

void do_top(char *arg)
{
	int number, lcount;

	if ((sscanf(arg, "%d %d", &number, &lcount) != 2) || (number < 1) || (lcount < 0)) {
		send_error("incorrect arguments");
		check_wccount();
		return;
	}
		
	/* md_top takes unsigned int's as args, so they get converted for us. */
	maildrop->md_top(number, lcount);
}

void do_uidl(char *arg) {
	unsigned int number = 1;

	if (arg[0]) {
		if ((sscanf(arg, "%u", &number) != 1) || number == 0) {
			send_error("incorrect argument");
			check_wccount();
		} else
			md_uidl(number, maildrop->md_md5_uidl_message);
	} else
		md_uidl(0, maildrop->md_md5_uidl_message);
}

void sig_handler(int number) {
	if (number == SIGALRM)
#ifdef LOG_EXTEND
		pop_log(pop_priority, "autologout time elapsed - %.384s",
			ahname);
#else
		pop_log(pop_priority, "autologout time elapsed");
#endif
	exit(1);
}

#ifdef NONIPVIRTUALS
void do_nonip(void) {
	char *tmpvname;

	tmpvname = strchr(username, '%');
	if (tmpvname == NULL)
		tmpvname = strchr(username, '@');
	if (tmpvname != NULL)
		*tmpvname = 0;
	if (tmpvname != NULL) {
		if (!allownonip)
			pop_log(pop_priority, "unallowed non-IP based virtual "
				"hosting request rejected");
		else
			if (parse_options(gargc, gargv, tmpvname + 1) < 0) {
				send_error("fatal error");
				exit(1);
			}
	}
}
#endif
#ifdef MAPPING
void do_mapusername(void) {
	int tmp;

	mapusername[0] = 0;
	if (domapping) {
		if (preauth_map_format[0]) {
/* map_find_preauth_mapping() doesn't overflow mapusername - 
   sizeof(mapusername) = MAXARGLN + 1 */
/* RequiredMapping is ignored here - preauth mapping always return mapped
   username. */
			pop_log_dbg(pop_priority, "Looking up pre-auth "
				"suffix for %.40s, %.40s\n", username,
				preauth_map_format);
			if (((tmp = map_find_preauth_mapping(preauth_map_format,
				mapusername)) < 0)) {
# ifdef LOG_EXTEND
				pop_log(pop_priority, "authentication failed: "
					"can't map user name: %.40s - %.384s",
					username, ahname);
# else
				pop_log(pop_priority, "authentication failed: "
					"can't map user name: %.40s",
					username);
# endif
				sleep(1);
				send_error("authentication failed");
				exit(1);
			}
			if (strlen(mapusername) >= sizeof(mapusername)) {
				pop_log(pop_priority, "map: unexpected "
					"error - contact author "
					"of program!!!");
				send_error("fatal error");
				exit(1);
			}
			pop_log_dbg(pop_priority, "mapped username: %.40s",
				mapusername);
		} else {
/* map_finduser() doesn't overflow mapusername - 
   sizeof(mapusername) = MAXARGLN + 1 */
			if (((tmp = map_finduser(sp_mapfile, username, 
				mapusername)) < 0) && (reqmapping)) {
# ifdef LOG_EXTEND
				pop_log(pop_priority, "authentication failed: "
					"can't map user name: %.40s - %.384s",
					username, ahname);
# else
				pop_log(pop_priority, "authentication failed: "
					"can't map user name: %.40s",
					username);
# endif
				sleep(1);
				send_error("authentication failed");
				exit(1);
			}
			if (tmp < 0) { /* && !reqmapping */
				domapping = 0;
				strcpy(mapusername, username);/* no overflow */
				/* it's for '%m' substitution in expand_dir */
			} else
				if (strlen(mapusername) >= sizeof(mapusername)) {
					pop_log(pop_priority, "map: unexpected "
						"error - contact author "
						"of program!!!");
					send_error("fatal error");
					exit(1);
				}
		}
	}

}
#endif

void do_authentication(char *pass) {

#ifdef SPTLS
	if (requiretls && ((encenabled & 2) == 0)) {
		send_error("you must establish TLS connection first");
		check_wccount();
		return;
	}
#endif
	if (username[0] == 0) {
		memset(pass, 0, strlen(pass));
		send_error("unknown username");
		check_wccount();
		return;
	}
	if (pass[0] == 0) {
		send_error("password required");
		check_wccount();
		return;
	}
#ifdef NONIPVIRTUALS
	do_nonip();
#endif
#ifdef MAPPING
	do_mapusername();
	write_general_auth(AUTH_USER, username, pass, mapusername);
#else
	write_general_auth(AUTH_USER, username, pass);
#endif
/* there is no specific USER/PASS authentication parameters */
	memset(pass, 0, strlen(pass));
	auth_finished = 1;
	return;
}

#ifdef APOP
void do_apop(char *arg) {
	char *tmp;
	int tmp2;
	
	if (!allowapop) {
		memset(arg, 0, strlen(arg));
		send_error("you are not allowed to authenticate yourself "
			"through APOP");
		check_wccount();
		return;
	}
# ifdef SPTLS
	if (requiretls && ((encenabled & 2) == 0)) {
		send_error("you must establish TLS connection first");
		check_wccount();
		return;
	}
# endif
	if ((tmp = strchr(arg, ' ')) == NULL) {
		memset(arg, 0, strlen(arg));
		send_error("argument missing");
		check_wccount();
		return;
	}
	*tmp = 0;
	tmp++;
	if (strlen(arg) >= sizeof(username)) {
		memset(tmp, 0, strlen(tmp));
		send_error("username too long");
		check_wccount();
		return;
	}
	strncpy(username, arg, sizeof(username)); /* Pad with NULs */
	
	if (strlen(tmp) != 32) {
		memset(tmp, 0, strlen(tmp));
		send_error("wrong digest length");
		check_wccount();
		return;
	}
	for (tmp2 = 0; tmp2 < 16; tmp2++)
		if ((((tmp[tmp2 * 2] > 'f') || (tmp[tmp2 * 2] < 'a')) &&
			(!isdigit(tmp[tmp2 * 2]))) ||
			(((tmp[(tmp2 * 2) + 1] > 'f') ||
			(tmp[(tmp2 * 2) + 1] < 'a')) &&
			(!isdigit(tmp[(tmp2 * 2) + 1])))) {
		    	memset(tmp, 0, strlen(tmp));
			send_error("wrong character in digest");
			check_wccount();
			return;
		}
# ifdef NONIPVIRTUALS
	do_nonip();
# endif
# ifdef MAPPING
	do_mapusername();
	write_general_auth(AUTH_APOP, username, tmp, mapusername);
# else
	write_general_auth(AUTH_APOP, username, tmp);
# endif
	if (write_string(tunnel[1], apoptimestamp) < 0)
		write_auth_error();
	memset(tmp, 0, strlen(tmp));
	auth_finished = 1;
	return;
}
#endif /* APOP */


void child_corrupted(void) {
/* This function works with root privileges */
	pop_log(pop_priority, "child process corrupted");
	exit(1);
}


#ifdef STANDALONE
int do_session(int argc, char **argv)
#else
int main(int argc, char **argv)
#endif
{
	const struct s_cmd *act_cmd;
#ifdef SPIPV6
	union sp_sockaddr peeraddr;
#else
	struct sockaddr_in peeraddr;
#endif
	socklen_t addrln;
	struct rlimit corelimit = {0, 0};
	struct passwd *spop3d, *userentry;
	int tmp, rootrequired = 1;
	gid_t tmpgid;
#ifdef SPSASL
	int sasl_username;
#endif

	/* That below works as root */
	signal(SIGHUP, SIG_IGN);
	signal(SIGINT, sig_handler);
	signal(SIGQUIT, sig_handler);
	signal(SIGILL, sig_handler);
	signal(SIGABRT, sig_handler);
	signal(SIGFPE, sig_handler);
	signal(SIGSEGV, sig_handler);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGALRM, sig_handler);
	signal(SIGTERM, sig_handler);
	signal(SIGUSR1, sig_handler);
	signal(SIGUSR2, sig_handler);
	signal(SIGCHLD, SIG_IGN);
	
	gargc = argc;
	gargv = argv;
	memset(username, 0, sizeof(username));
	memset(password, 0, sizeof(username));
#ifdef MAPPING
	memset(mapusername, 0, sizeof(mapusername));
#endif
	alarm(0);
	umask(0077);
	chdir("/");
	
#ifndef STANDALONE
	pop_openlog();
	if (atexit(spop3d_done) < 0)
		exit(1);
#endif
	if (setrlimit(RLIMIT_CORE, &corelimit) < 0) {
		pop_error("setrlimit");
		exit(1);
	}
#ifdef SPIPV6
	addrln = sizeof(union sp_sockaddr);
	if (getpeername(0, (struct sockaddr *)&peeraddr.saddr_in6,
		&addrln) < 0) {
#else
	addrln = sizeof(struct sockaddr_in);
	if (getpeername(0, (struct sockaddr *)&peeraddr, &addrln) < 0) {
#endif
		pop_error("getpeername");
		exit(1);
	}
#ifdef SPIPV6
	if ((peeraddr.saddr_in6.sin6_family != AF_INET) &&
		(peeraddr.saddr_in6.sin6_family != AF_INET6)) {
#else
	if (peeraddr.sin_family != AF_INET) {
#endif
		pop_log(pop_priority, "peer address is not an IP address");
		exit(1);
	}
#if defined(LOG_EXTEND) || defined(LOG_CONNECT) || defined(STANDALONE)
	make_ahname(&peeraddr, ahname);
#endif /* LOG_EXTEND || LOG_CONNECT || STANDALONE */

#ifdef SPENCRYPTION
	if (open_encpipes() < 0)
		exit(1);
#endif

	if (pipe(tunnel) < 0) {
		pop_error("pipe");
		exit(1);
	}
#ifdef SPSASL
	sasl_pipe();
#endif
	switch (fork()) {
		case -1:
			pop_error("fork");
			exit(1);
		case 0:
			close(tunnel[0]);
			tunnel[0] = -1;
#ifdef SPENCRYPTION
			server_encpipes();
#endif

#ifdef SPSASL
			sasl_client_pipe();
#endif
			if ((spop3d = getpwnam(POPUSER)) == NULL) {
				pop_log(pop_priority, "user " POPUSER
					" not found");
				pop_error("getpwnam");
				exit(1);
			}
			if (set_privileges(spop3d->pw_uid, spop3d->pw_gid) < 0)
				exit(1);
/* Code below works as spop3d user */
			strcpy(maildrop_name, DEFMAILDROPNAME);
			strcpy(maildrop_type, DEFMAILDROPTYPE);
#ifdef BULLETINS
			strcpy(userbullfile, USERBULL);
			strcpy(bulldir, BULLDIR);
#endif
#ifdef SPTLS
			cafile[0] = capath[0] = certfile[0] = keyfile[0] =
				dh512paramfile[0] = dh1024paramfile[0] =
				cipherlist[0] = rand_source[0] = 0;
#endif
#ifdef DPUID
			dpuid_db[0] = dpuid_user[0] = dpuid_log[0] = 0;
#endif
#ifdef LOG_EXTEND
			authlog[0] = 0;
#endif
#ifdef MAPPING
			preauth_map_format[0] = 0;
#endif
#ifdef MD5PASSWORD
			sppassword[0] = 0;
#endif
#ifdef SPSASL
			sasl_mechanism[0] = saslrealm[0] = 0;
#endif
#ifdef NONIPVIRTUALS
			if (parse_options(argc, argv, NULL) < 0)
#else
			if (parse_options(argc, argv) < 0)
#endif
				exit(1);

			write_logpriority();

#ifdef SPSASL

# ifndef STANDALONE
			if (allowsasl)
				init_sasl();
# endif
			if (allowsasl) {
				new_connection_sasl();
				setprops_sasl();
			}
#endif

#ifdef SPTLS
			if (enabletls)
				tls_enable(0);
#endif
#ifdef APOP
			if (allowapop) {
				if (apopservername[0] == 0)
					if (gethostname(apopservername,
						sizeof(apopservername)) < 0)
						exit(1);
				snprintf(apoptimestamp, sizeof(apoptimestamp),
					 "<%lu.%lu@%.256s>",
					 (unsigned long int)getpid(),
					 (unsigned long int)time(NULL),
					 apopservername);
				send_ok("%.100s %.400s", SERVER_GREETING,
					apoptimestamp);
			} else
				send_ok("%.100s", SERVER_GREETING);
#else
			send_ok(SERVER_GREETING);
#endif
			if (buf_flush() < 0)
				exit(1);
			connection_state = AUTH_STATE;
			while (!auth_finished) {
				read_command(cmd);
				if ((act_cmd = cmd_lookup(cmds[\
					connection_state], cmd)) == NULL) {
					send_error("unknown command");
					if (buf_flush() < 0)
						exit(1);
					check_wccount();
					continue;
				}
				if (strlen(cmd) == strlen(act_cmd->name))
					act_cmd->handler(cmd +
						strlen(act_cmd->name));
				else
					act_cmd->handler(cmd +
						strlen(act_cmd->name) + 1);
				if (buf_flush() < 0)
					exit(1);
			}
			close(tunnel[1]);
			tunnel[1] = -1;
#ifdef SPENCRYPTION
			if (encenabled)
				encryption_server();
#endif
			exit(0);
		default: ;
	} /* Code below reads configuration from child process.
	     Code below works with root privileges */
	close(tunnel[1]);
	tunnel[1] = -1;
	read_logpriority();
#if defined(LOG_CONNECT) || defined(STANDALONE)
	pop_log(pop_priority, "connect from %.384s", ahname);
#endif

#ifdef SPENCRYPTION
	client_encpipes();
# ifdef SPSASL
#  ifdef STANDALONE
	done_sasl();
#  endif
	sasl_server_pipe();
# endif
# if defined(SPTLS) && defined(STANDALONE)
	tls_done_serverengine();
# endif
#endif
#ifdef MAPPING
	memset(mapusername, 0, sizeof(mapusername));
#endif
	read_general_auth();
	switch (auth_method) {
		case AUTH_USER:
			break; /* nothing specific */
#ifdef APOP
		case AUTH_APOP:
			if (read_string(tunnel[0], apoptimestamp,
				sizeof(apoptimestamp)) < 0)
				child_corrupted();
# ifdef USERPASSWD
			allowuserpasswd = requserpasswd = 0;
# endif
			break;
#endif
#ifdef SPSASL
		case AUTH_SASL:
			if (read_loop(tunnel[0], &sasl_username,
				sizeof(sasl_username)) !=
				sizeof(sasl_username))
				child_corrupted();
# ifdef USERPASSWD
			allowuserpasswd = requserpasswd = 0;
# endif
			break;
#endif
		default:
				child_corrupted();
	} /* switch (auth_method) */
#ifdef SPENCRYPTION
# ifdef SPTLS
	tmp = 2;
# else
	tmp = 0;
# endif
	if (encenabled & (~tmp))
		child_corrupted();
	if (encenabled)
		enctrans = 1;
	if ((!encenabled) && (auth_method != AUTH_SASL))
		close_encpipes();
#endif /* SPENCRYPTION */
	check_logpriority(logpriority);
	auth_successful = -1;
#ifdef SPSASL
	if (auth_method != AUTH_SASL)
		sasl_close_pipe();
	if ((auth_method == AUTH_SASL) &&
	    (sasl_username == 0)) {
		pop_log(pop_priority, "authentication failed: SASL "
			"(unknown username)");
		sleep(1);
		send_error("authentication failed");
		exit(1);
	}
	if ((auth_method == AUTH_SASL) && (sasl_username != 1) &&
		(sasl_username != 2) && (sasl_username != (2 | 4)))
		child_corrupted();
	if (auth_method == AUTH_SASL) {
		if (sasl_username & 4)
			encenabled = enctrans = 1;
		sasl_username &= 3;
	}
	/* if auth_method == AUTH_SASL, sasl_username = 1 we know username,
	   but we still aren't authenticated
	   if auth_method == AUTH_SASL, sasl_username = 2 we know username
	   and we are authenticated through some SASL (EXTERNAL and maybe
	   other) mechanism */
#endif

#ifdef MAPPING
	if (domapping)
		userentry = spgetpwnam(mapusername, &rootrequired);
	else
		userentry = spgetpwnam(username, &rootrequired);
#else
		userentry = spgetpwnam(username, &rootrequired);
#endif

	if (userentry == NULL) {
		no_such_user();
		sleep(1);
		send_error("authentication failed");				
		exit(1);
	}

#ifdef APOP
	apop_secret[0] = 0;
#endif
	if (auth_method != AUTH_USER)
		rootrequired = 0;
#ifdef USERPASSWD
	if (requserpasswd) {
			rootrequired = 0; /* we must use
			password from ~/.spop3d */
			allowuserpasswd = 1;
	}
#endif
	if ((rootrequired) && (auth_method == AUTH_USER)) {
#ifdef MAPPING
		if (domapping)
			tmp = sp_authenticate_user(mapusername, 
				password);
		else
			tmp = sp_authenticate_user(username, password);
#else /* MAPPING */
		tmp = sp_authenticate_user(username, password);
#endif /* MAPPING */
		auth_successful = tmp;
#ifdef USERPASSWD
		if (allowuserpasswd)
			rootrequired = 0; 
			/* may be in ~/.spop3d */
#endif
	}
	if (rootrequired)
		after_authentication(auth_successful, NULL);

#ifdef ALLOWROOTLOGIN
	if ((userentry->pw_uid == 0) && (!allow_root)) {
#else
	if (userentry->pw_uid == 0) {
#endif
#ifdef LOG_EXTEND
		pop_log(pop_priority, "root login not allowed - %.384s",
			ahname);
#else
		pop_log(pop_priority, "root login not allowed");
#endif
		sleep(1);
		send_error("authentication failed");
		exit(1);
	}
#ifdef LOG_EXTEND
	if (authlog[0])
		authlog_openlog(authlog);
#endif /* LOG_EXTEND */
	if (changegid)
		tmpgid = userentry->pw_gid;
	else
		tmpgid = getgid();
	if (set_privileges(userentry->pw_uid, tmpgid) < 0) {
		send_error("fatal error");
		exit(1);
	}
/* Code below works with user privileges */
#ifdef DPUID
	if (dpuid_enabled) {
# ifdef MAPPING
		tmp = (domapping ? (dpuid_set(mapusername)) :
		       (dpuid_set(username)));
# else
		tmp = dpuid_set(username);
# endif
		if (tmp < 0) {
			after_authentication(-1, dpuid_log);
			exit(1); /* shouldn't happen */
		}
# ifdef USERPASSWD
	requserpasswd = allowuserpasswd = 0;
# endif
	}
#endif /* DPUID */

#ifdef USERCONFIG
	if (useroverride)
		parse_user_cfg(userentry->pw_dir);
#endif

#ifdef USERPASSWD
	if (auth_method == AUTH_USER) {
		if ((requserpasswd) && (sppassword[0] == 0))
			after_authentication(-1, NULL);
		if ((sppassword[0] == 0) && (auth_successful < 0))
			after_authentication(-1, NULL);
		if (sppassword[0])
			auth_successful = -1;
	}
#endif

#ifdef MD5PASSWORD
	if ((!rootrequired) && (auth_method == AUTH_USER))
# ifdef DPUID
		after_authentication((auth_successful < 0) ?
			verify_password(password) : 0,
			dpuid_log);
# else
		after_authentication((auth_successful < 0) ?
			verify_password(password) : 0, NULL);
# endif
#endif /* MD5PASSWORD */

#ifdef APOP			
	if (auth_method == AUTH_APOP) {
		if (apop_secret[0] == 0) {
# ifdef LOG_EXTEND
			pop_log(pop_priority, "can't find APOP secret for "
				"user %.40s - %.384s", username, ahname);
# else
			pop_log(pop_priority, "can't find APOP secret for "
				"user %.40s", username);
# endif
			sleep(1);
			send_error("authentication failed");
			exit(1);
		}
# ifdef DPUID
		after_apop(apop_authenticate(username,
			apoptimestamp, password), dpuid_log);
# else
		after_apop(apop_authenticate(username,
			apoptimestamp, password), NULL);
# endif
	}
	memset(apop_secret, 0, sizeof(apop_secret));
#endif /* APOP */

#ifdef SPSASL
	if (auth_method == AUTH_SASL) {
		if (sasl_secret_len == 0) {
# ifdef LOGEXTEND
			pop_log(pop_priority, "can't find SASL secret for "
				"user %.40s - %.384s", username, ahname);
# else
			pop_log(pop_priority, "can't find SASL secret for "
				"user %.40s", username);
# endif
		}
# ifdef DPUID
		after_sasl((sasl_username == 1) ? sasl_authenticate() : 0,
			dpuid_log);
# else
		after_sasl((sasl_username == 1) ? sasl_authenticate() : 0,
			NULL);
		sasl_close_pipe();
# endif
	}
	memset(sasl_secret, 0, sizeof(sasl_secret));
#endif /* SASL */
	memset(password, 0, strlen(password));
	close(tunnel[0]);
	tunnel[0] = -1;
	if (expand_dir(maildrop_name, userentry->pw_dir, sizeof(maildrop_name))
		< 0) {
		pop_log(pop_priority, "error while expanding maildrop name");
		sleep(1);
		send_error("authentication failed");
		exit(1);
	}
	maildrop = find_maildrop(maildrop_type);
	maildrop->md_init();
	releasemaildrop = 1;
#ifdef BULLETINS
	add_bulletins(userentry->pw_dir);
#endif
	connection_state = TRANSACTION_STATE;
#ifdef LOG_EXTEND
	authlog_log();
	authlog_closelog();
#endif /* LOG_EXTEND */
	send_ok("authentication successful");
	if (buf_flush() < 0)
		exit(1);
	while (1) {
		read_command(cmd);
		if ((act_cmd = cmd_lookup(cmds[connection_state], cmd))
			== NULL) {
			send_error("unknown command");
			if (buf_flush() < 0)
				exit(1);
			check_wccount();
			continue;
		}
		if (strlen(cmd) == strlen(act_cmd->name))
			act_cmd->handler(cmd + strlen(act_cmd->name));
		else
			act_cmd->handler(cmd + strlen(act_cmd->name) + 1);
		if (buf_flush() < 0)
			exit(1);
	}
}
