/* This part was written by Bodo Moeller <bodo@openssl.org>.
 * Used with permission. */

#include "includes.h"

#include "md5.h"

unsigned const char cov_2char[64]={
	/* from crypto/des/fcrypt.c */
	0x2E,0x2F,0x30,0x31,0x32,0x33,0x34,0x35,
	0x36,0x37,0x38,0x39,0x41,0x42,0x43,0x44,
	0x45,0x46,0x47,0x48,0x49,0x4A,0x4B,0x4C,
	0x4D,0x4E,0x4F,0x50,0x51,0x52,0x53,0x54,
	0x55,0x56,0x57,0x58,0x59,0x5A,0x61,0x62,
	0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6A,
	0x6B,0x6C,0x6D,0x6E,0x6F,0x70,0x71,0x72,
	0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A
};

char *md5_crypt(const char *passwd, const char *salt)
	{
	static char out_buf[3 + 9 + 24 + 2]; /* "$1$..salt..$.......md5hash..........\0" */
	unsigned char buf[16];
	char *salt_out, *tmp;
	int n, i;
	struct md5_ctx md;
	size_t passwd_len, salt_len;

	passwd_len = strlen(passwd);
	if (strncmp(salt, "$1$", 3) != 0)
		return NULL;
	salt += 3;
	if ((tmp = strchr(salt, '$')) == NULL)
		return NULL;
	if ((salt_len = (tmp - salt)) > 8)
		return NULL;
	strcpy(out_buf, "$1$");
	salt_out = out_buf + 3;
	memcpy(salt_out, salt, salt_len);
	salt_out[salt_len] = 0;
	
	md5_init_ctx(&md);
	md5_process_bytes(passwd, passwd_len, &md);
	md5_process_bytes("$1$", 3, &md);
	md5_process_bytes(salt_out, salt_len, &md);
	
	 {
		struct md5_ctx md2;

		md5_init_ctx(&md2);
		md5_process_bytes(passwd, passwd_len, &md2);
		md5_process_bytes(salt_out, salt_len, &md2);
		md5_process_bytes(passwd, passwd_len, &md2);
		md5_finish_ctx(&md2, buf);
		memset(&md2, 0, sizeof(md2));
	 }

	for (i = passwd_len; i > sizeof buf; i -= sizeof buf)
		md5_process_bytes(buf, sizeof buf, &md);
	md5_process_bytes(buf, i, &md);
	
	n = passwd_len;
	while (n)
		{
		md5_process_bytes((n & 1) ? "\0" : passwd, 1, &md);
		n >>= 1;
		}
	md5_finish_ctx(&md, buf);

	for (i = 0; i < 1000; i++)
		{
		struct md5_ctx md2;

		md5_init_ctx(&md2);
		md5_process_bytes((i & 1) ? (unsigned char *) passwd : buf,
		                 (i & 1) ? passwd_len : sizeof buf,
				  &md2);
		if (i % 3)
			md5_process_bytes(salt_out, salt_len, &md2);
		if (i % 7)
			md5_process_bytes(passwd, passwd_len, &md2);
		md5_process_bytes((i & 1) ? buf : (unsigned char *) passwd,
		                 (i & 1) ? sizeof buf : passwd_len,
				  &md2);
		md5_finish_ctx(&md2, buf);
		memset(&md2, 0, sizeof(md2));
		}
	
	 {
		/* transform buf into output string */
	
		unsigned char buf_perm[sizeof buf];
		int dest, source;
		char *output;

		/* silly output permutation */
		for (dest = 0, source = 0; dest < 14; dest++,
			source = (source + 6) % 17)
			buf_perm[dest] = buf[source];
		buf_perm[14] = buf[5];
		buf_perm[15] = buf[11];
		
		output = salt_out + salt_len;
		
		*output++ = '$';

		for (i = 0; i < 15; i += 3)
			{
			*output++ = cov_2char[buf_perm[i+2] & 0x3f];
			*output++ = cov_2char[((buf_perm[i+1] & 0xf) << 2) |
				                  (buf_perm[i+2] >> 6)];
			*output++ = cov_2char[((buf_perm[i] & 3) << 4) |
				                  (buf_perm[i+1] >> 4)];
			*output++ = cov_2char[buf_perm[i] >> 2];
			}
		*output++ = cov_2char[buf_perm[i] & 0x3f];
		*output++ = cov_2char[buf_perm[i] >> 6];
		*output = 0;
		memset(buf_perm, 0, sizeof(buf_perm));
	 }

	memset(buf, 0, sizeof(buf));
	return out_buf;
	}
