static const char rcsid[] = "$Id: pop_auth.c,v 1.6 2000/09/29 10:31:54 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include <pwd.h>
#include <sys/file.h>
#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <errno.h>
#include "const.h"
#include "maildrop.h"
#include "fdfgets.h"

int main(int argc, char **argv) {
	int fd, md_set = 0, fret;
	unsigned int linenr;
	ssize_t tmp;
	struct passwd *pwentry;
	char username[MAXARGLN + 1];
	char *pass;
	char passcopy[MAXARGLN + 1];
	char txtbuff[256], buf[128];
	struct rlimit corelimit = {0, 0};
	char cfgfile[PATH_MAX];
	struct stat stbuf;
	char *tmp2;
	char tmpmaildrop_type[MAXMDTYPENAMELENGTH], tmpmaildrop_name[PATH_MAX];
#ifdef USERPASSWD
	char tmpsppassword[PASSWORDLN];
	int pw_set = 0;
#endif
	static char digits[] = "0123456789abcdef";
	
	if (setrlimit(RLIMIT_CORE, &corelimit) < 0) {
		perror("setrlimit");
		exit(1);
	}
	if ((pwentry = getpwuid(getuid())) == NULL) {
		fprintf(stderr, "can't find user with UID: %u\n", getuid());
		exit(1);
	}
	username[0] = 0;
	strncat(username, pwentry->pw_name, sizeof(username) - 1);
	if (strlen(pwentry->pw_name) >= sizeof(username))
		fprintf(stderr, "Warning: username truncated\n");
	snprintf(txtbuff, sizeof(txtbuff),
		"Enter NEW APOP secret for user %.40s: ", username);
	pass = getpass(txtbuff);
	passcopy[0] = 0;
	strncat(passcopy, pass, sizeof(passcopy) - 1);
	memset(pass, 0, strlen(pass));
	snprintf(txtbuff, sizeof(txtbuff),
		"Reenter NEW APOP secret for user %.40s: ", username);
	pass = getpass(txtbuff);
	fret = strcmp(pass, passcopy);
	memset(pass, 0, strlen(pass));
	if (fret != 0) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "APOP secrets don't match\n");
		exit(1);
	}
	if (stat(pwentry->pw_dir, &stbuf) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "can't stat user home directory: %.1024s\n",
			pwentry->pw_dir);
		perror("stat");
		exit(1);
	}
	if (stbuf.st_mode & 2) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "user home directory is world writable\n");
		exit(1);
	}
	if ((strlen(pwentry->pw_dir) + strlen(USERCFG) + 2) > sizeof(cfgfile)) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "home directory name too long\n");
		exit(1);
	}
	strcpy(cfgfile, pwentry->pw_dir);
	strcat(cfgfile, "/");
	strcat(cfgfile, USERCFG);
	if ((fd = open(cfgfile, O_RDWR | O_CREAT, 0600)) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		fprintf(stderr, "can't open user config file\n");
    		perror("open");
		exit(1);
	}
	if (fstat(fd, &stbuf) < 0) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		fprintf(stderr, "can't stat user config file\n");
		perror("fstat");
		exit(1);
	}
	if (!S_ISREG(stbuf.st_mode)) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		fprintf(stderr, "user config file is not regular file\n");
		exit(1);
	}
	if (stbuf.st_mode & 0177) {
		memset(passcopy, 0, strlen(passcopy));
		close(fd);
		fprintf(stderr, "user config file has wrong mode\n");
		exit(1);
	}
	fd_initfgets();
	linenr = 0;
	while (1) {
    		linenr++;
		if ((tmp = fd_fgets(buf, sizeof(buf), fd, NULL)) < 0) {
			memset(passcopy, 0, strlen(passcopy));
#ifdef USERPASSWD
			memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
			memset(buf, 0, sizeof(buf));
			fd_cleanup();
    			close(fd);
			fprintf(stderr, "can't read user config file\n");
			perror("read");
			exit(1);
		}
		if (tmp == 0)
			break;
		if ((tmp == sizeof(buf)) && (buf[sizeof(buf) - 1] != '\n')) {
#ifdef USERPASSWD
			memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
			memset(passcopy, 0, strlen(passcopy));
			memset(buf, 0, sizeof(buf));
			fd_cleanup();
			close(fd);
			fprintf(stderr,
			    "line too long in user config file, line %u\n",
			    linenr);
			exit(1);
		}
		if ((tmp == 1) && (buf[0] == '\n'))
			continue;
		if (buf[tmp - 1] == '\n')
			buf[tmp - 1] = 0;
		else
			buf[tmp] = 0;
		strtok(buf, " \t");
		if (strcasecmp(buf, "maildrop") == 0) {
			if (md_set == 1) {
#ifdef USERPASSWD
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "maildrop already set in user "
					"config file, line: %u\n", linenr);
				exit(1);
			}
			if ((tmp2 = strtok(NULL, " \t")) == NULL) {
#ifdef USERPASSWD
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				close(fd);
				fd_cleanup();
				fprintf(stderr, "argument missing in maildrop "
					"declaration in user config file, "
					"line: %u\n", linenr);
				exit(1);
			}
			if (strlen(tmp2) >= sizeof(tmpmaildrop_name)) {
#ifdef USERPASSWD
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				close(fd);
				fd_cleanup();
				fprintf(stderr, "maildrop file name in user "
					"config file is too long, line: %u\n", 
					linenr);
				exit(1);
			}
			strcpy(tmpmaildrop_name, tmp2);
			if ((tmp2 = strtok(NULL, " \t")) == NULL) {
#ifdef USERPASSWD
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "argument missing in maildrop "
					"declaration in user config file, "
					"line: %u\n", linenr);
				exit(1);
			}
			if (strlen(tmp2) >= sizeof(tmpmaildrop_type)) {
#ifdef USERPASSWD
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "maildrop type in user config "
					"file is too long, line: %u\n", linenr);
				exit(1);
			}
#ifdef MDMAILBOX
			if (strcasecmp(tmp2, "mailbox") != 0)
#endif
#ifdef MDMAILDIR
			if (strcasecmp(tmp2, "maildir") != 0)
#endif
#ifdef MDCRLFMAILDIR
			if (strcasecmp(tmp2, "crlfmaildir") != 0)
#endif	    
			{
#ifdef USERPASSWD
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "no such maildrop type: %.40s, "
					"line: %u\n", tmp2, linenr);
    				exit(1);
			}
			strcpy(tmpmaildrop_type, tmp2);
			md_set = 1;
			continue;
		}
#ifdef USERPASSWD
		if (strcasecmp(buf, "Password") == 0) {
			if (pw_set == 1) {
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
				memset(passcopy, 0, strlen(passcopy));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "password already set in user "
					"config file, line: %u\n", linenr);
				exit(1);
			}
			if ((tmp2 = strtok(NULL, " \t")) == NULL) {
				memset(passcopy, 0, strlen(passcopy));
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "argument missing in maildrop "
					"declaration in password, line: %u\n",
					linenr);
				exit(1);
			}
			if (strlen(tmp2) >= sizeof(tmpsppassword)) {
				memset(passcopy, 0, strlen(passcopy));
				memset(tmpsppassword, 0, sizeof(tmpsppassword));
				memset(buf, 0, sizeof(buf));
				fd_cleanup();
				close(fd);
				fprintf(stderr, "password in user config "
					"file is too long, line: %u\n",
					linenr);
				exit(1);
			}
			strcpy(tmpsppassword, tmp2);
			pw_set = 1;
			continue;
		}
#endif
		if (strcasecmp(buf, "APOPSecret") == 0) {
			memset(buf, 0, sizeof(buf));
			continue;
		}
		memset(passcopy, 0, strlen(passcopy));
#ifdef USERPASSWD
		memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
		memset(buf, 0, sizeof(buf));
		fd_cleanup();
		close(fd);
		fprintf(stderr, "unknown option name, line: %u\n", linenr);
		exit(1);
	}
	memset(buf, 0, sizeof(buf));
	fd_cleanup();
	if (lseek(fd, 0, SEEK_SET) < 0) {
		memset(passcopy, 0, strlen(passcopy));
#ifdef USERPASSWD
		memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
		close(fd);
		perror("lseek");
		exit(1);
	}
	if (md_set == 1) {
		snprintf(txtbuff, sizeof(txtbuff), "MailDrop %.100s %.100s\n",
			tmpmaildrop_name, tmpmaildrop_type);
		if (write(fd, txtbuff, strlen(txtbuff)) < 0) {
#ifdef USERPASSWD
			memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
			memset(passcopy, 0, strlen(passcopy));
			close(fd);
			perror("write");
			exit(1);
		}
	}
#ifdef USERPASSWD
	memset(txtbuff, 0, sizeof(txtbuff));
	if (pw_set == 1) {
		snprintf(txtbuff, sizeof(txtbuff), "Password %.128s\n",
			tmpsppassword);
		memset(tmpsppassword, 0, sizeof(tmpsppassword));
		if (write(fd, txtbuff, strlen(txtbuff)) < 0) {
			memset(passcopy, 0, strlen(passcopy));
			close(fd);
			perror("write");
			exit(1);
		}
		memset(txtbuff, 0, sizeof(txtbuff));
	}
#else
	memset(txtbuff, 0, sizeof(txtbuff));
#endif

	strcpy(txtbuff, "APOPSecret ");
	tmp2 = txtbuff + strlen(txtbuff);
	for (tmp = 0; tmp < strlen(passcopy); tmp++) {
		passcopy[tmp] ^= 0xff;
		tmp2[tmp * 2] = digits[(passcopy[tmp] >> 4) & 0x0f];
		tmp2[(tmp * 2) + 1] = digits[passcopy[tmp] & 0x0f];				
	}
	memset(passcopy, 0, strlen(passcopy));
	strcat(txtbuff, "\n");
	if (write(fd, txtbuff, strlen(txtbuff)) < 0) {
		memset(txtbuff, 0, strlen(txtbuff));
		close(fd);
		perror("write");
		exit(1);
	}
	memset(txtbuff, 0, strlen(txtbuff));
	if ((tmp = lseek(fd, 0, SEEK_CUR)) < 0) {
		close(fd);
		perror("lseek");
		exit(1);
	}
	if (ftruncate(fd, tmp) < 0) {
		close(fd);
		perror("ftruncate");
		exit(1);
	}
	if (close(fd) < 0) {
		perror("close");
		exit(1);
	}
	exit(0);
}
