static const char rcsid[] = "$Id: spdbm.c,v 1.6 2000/09/03 13:28:47 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include "mydbm.h"

#ifdef MD5PASSWORD
# ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
# else
#  ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#  else
#   include <time.h>
#  endif
# endif
# include "md5.h"
# include "md5_crypt.h"
#endif
extern unsigned const char cov_2char[];
#ifdef SPSASL
# include <sasl.h>
# include <saslutil.h>
struct secret_str {
	int len;
	char secret[SASLSECRETLN];
};
struct secret_str spdbm_sasl_secret;
sasl_conn_t *connection;
#endif

#define MYDBM_SPDBM

#include "mydbm.c" /* ugly */

extern int optopt, opterr, optind;
extern char *optarg;

void usage(char *arg) {
	fprintf(stderr, "\n"
"Usage: %s -f dbname [-m mapname] [-rcodl] [username]\n"
"  -f dbname               name of database (required)\n", arg);
#ifdef MAPPING
	fprintf(stderr,
"  -m mapname              add username:mapname entry to database\n");
#endif
#ifdef DPUID
	fprintf(stderr,
"  -p password             set password for username from dpuid domain\n");
# ifdef APOP
	fprintf(stderr,
"  -a apopsecret           set APOP secret for username from dpuid domain\n");
# endif
# ifdef SPSASL
	fprintf(stderr,
"  -s saslsecret           set SASL secret for username from dpuid domain\n"
"  -b realm                set SASL realm for created SASL secrets\n");
# endif
	fprintf(stderr,
"  -n maildrop_name        set maildrop_name for username from dpuid domain\n"
"  -t maildrop_type        set maildrop_type for username from dpuid domain\n"
"  -e                      password is already MD5-encrypted, "
"don't encrypt it\n");
# ifdef MAPPING
	fprintf(stderr,
"  -u                      set database type to dpuid, default is mapping\n");
# endif
#endif
	fprintf(stderr,
"  -r                      replace existing entries when adding\n"
"  -c                      create database if it doesn't exist\n"
"  -o                      create empty database regardless if one exists\n"
"  -d                      delete username from database\n"
"  -l                      list entries from database\n");
}

void store_field(char *buf, char *what, unsigned long int length,
	int *valuesize) {
	char *act = buf + (*valuesize);

	*((unsigned long int *) act) = length;
	memcpy(act + sizeof(unsigned long int), what, length);
	*valuesize += (length + sizeof(unsigned long int));
}

#ifdef MAPPING
void map_add_entries(char *dbfile, int replace, int create, char *username,
		     char *mapname) {
	MYDBM db;
	char buf[(2 * MAXARGLN) + 3]; /* username:mapname\n\0 */
	char value[MAXARGLN + sizeof(unsigned long int)];
	char *sep;
	int c, valuesize;
	
	if (mydbm_open(&db, dbfile, create) < 0)
		exit(1);
	if (username != NULL) {
		valuesize = 0;
		store_field(value, mapname, strlen(mapname), &valuesize);
		c = mydbm_store(&db, username, value, valuesize,
			replace);
		if (c < 0)
			exit(1);
		if (c == 1)
			fprintf(stderr, "entry %s exists in database but "
				"wasn't replaced\n", username);
		else
			fprintf(stderr, "added/replaced entry in database\n");
		mydbm_close(&db);
		return;
	}
	while (fgets(buf, sizeof(buf), stdin) != NULL) {
		if (strlen(buf) == (sizeof(buf) - 1)) {
			fprintf(stderr, "line too long\n");
			mydbm_close(&db);
			exit(1);
		}
		sep = buf + strlen(buf) - 1;
		if ((*sep) == '\n')
			(*sep) = 0;
		if ((sep = strchr(buf, ':')) == NULL) {
			fprintf(stderr, "missing separator in line: %s\n",
				buf);
			mydbm_close(&db);
			exit(1);
		}
		(*sep) = 0;
		sep++;
		if (strlen(buf) > MAXARGLN) {
			fprintf(stderr, "username: %s too long\n", buf);
			mydbm_close(&db);
			exit(1);
		}
		if (strlen(sep) > MAXARGLN) {
			fprintf(stderr, "mapname: %s too long\n", sep);
			mydbm_close(&db);
			exit(1);
		}
		valuesize = 0;
		store_field(value, sep, strlen(sep), &valuesize);
		c = mydbm_store(&db, buf, value, valuesize, replace);
		if (c < 0)
			exit(1);
		if (c == 1)
			fprintf(stderr, "entry %s exists in database "
				"but wasn't replaced\n", buf);
	}
	mydbm_close(&db);
}
#endif /* MAPPING */

void delete_entries(char *dbfile, char *username) {
	MYDBM db;
	char *sep;
	char buf[4 * (MAXARGLN + 1) + PATH_MAX + 1 + MAXMDTYPENAMELENGTH + 2];
	/* username:password:apop_secret:saslsecret:maildrop_name:
	   maildrop_type\n\0 */
	/* username:mapname\n\0 */
	
	if (mydbm_open(&db, dbfile, DB_WRITER) < 0)
		exit(1);
	if (username != NULL) {
		if (mydbm_delete(&db, username) < 0)
			exit(1);
		fprintf(stderr, "removed entry from database\n");
		mydbm_close(&db);
		return;
	}
	while (fgets(buf, sizeof(buf), stdin) != NULL) {
		if ((sep = strchr(buf, ':')) != NULL)
			(*sep) = 0;
		sep = buf + strlen(buf) - 1;
		if ((*sep) == '\n')
			(*sep) = 0;
		mydbm_delete(&db, buf); /* ignore errors */
		memset(buf, 0, sizeof(buf));
	}
	mydbm_close(&db);
}

void list_entries(char *dbfile) {
	MYDBM db;
	char key[MAXARGLN + 1];
	char value[MAXARGLN + PASSWORDLN + SASLSECRETLN + PATH_MAX + 1 + 
		MAXMDTYPENAMELENGTH + (5 * sizeof(unsigned long int))];
	/* password:apop_secret:sasl_secret:maildrop_name:maildrop_type\0 */
	/* mapname */
	int valuesize = sizeof(value);
	unsigned long int length;
	int entry;
	int nr;
	char *tmp;
	
	if (mydbm_open(&db, dbfile, DB_READER) < 0)
		exit(1);
	entry = mydbm_firstentry(&db, key, sizeof(key), value, &valuesize);
	while (entry == 1) {
		tmp = value;
		nr = 0;
		printf("%s:", key);
		while (valuesize >= sizeof(unsigned long int)) {
			length = *((unsigned long int *) tmp);
			tmp += sizeof(unsigned long int);
			valuesize -= sizeof(unsigned long int);
			if (length > valuesize) {
				fprintf(stderr, "\nentry damaged\n");
				memset(value, 0, sizeof(value));
				return;
			}
			/* APOP or SASL secret */
			if (((nr == 1) || (nr == 2)) && (length > 0))
				printf("*");
			else
				fwrite(tmp, 1, length, stdout);
			tmp += length;
			valuesize -= length;
			if (valuesize > 0)
				printf(":");
			nr++;

		}
		memset(value, 0, sizeof(value));
		if (valuesize != 0) {
			fprintf(stderr, "\nentry damaged\n");
			return;
		}
		printf("\n");
		valuesize = sizeof(value);
		entry = mydbm_nextentry(&db, key, sizeof(key), value,
			&valuesize);
	}
	mydbm_close(&db);
	if (entry < 0)
		exit(1);
}

#ifdef DPUID
char *generate_salt(char *buf) { /* buf should be a 13-byte buffer */
	struct md5_ctx ctx;
	char digest[16];
	int i;
	struct timeval a;
	clock_t b = clock();
	pid_t c = getpid(), d = getppid();
	uid_t e = getuid(), f = getgid();
	
# ifdef GETTIMEOFDAY_TZ
	gettimeofday(&a, NULL);
# else
	gettimeofday(&a);
# endif
	md5_init_ctx(&ctx);
	md5_process_bytes(&a, sizeof(a), &ctx);
	md5_process_bytes(&b, sizeof(b), &ctx);
	md5_process_bytes(&c, sizeof(c), &ctx);
	md5_process_bytes(&d, sizeof(d), &ctx);
	md5_process_bytes(&e, sizeof(e), &ctx);
	md5_process_bytes(&f, sizeof(f), &ctx);
	md5_process_bytes(buf, 13, &ctx);
	md5_finish_ctx(&ctx, digest);
	memset(&ctx, 0, sizeof(ctx));
	memset(buf, 0, 13);
	strcpy(buf, "$1$");
	for (i = 0; i < 8; i++)
		buf[i + 3] = cov_2char[digest[i] & 0x3f];
	strcat(buf, "$");
	return buf;
}

# ifdef SPSASL
void print_errors(char *name, int result) {
	fprintf(stderr, "%s: %s\n", name, sasl_errstring(result, NULL, NULL));
}

int spdbm_getsecret(void *context, const char *mechanism,
	const char *auth_identity, const char *realm, sasl_secret_t **secret) {
	return SASL_NOUSER;
}

int spdbm_log(void *context, int priority, const char *message) {
	switch (priority) {
		case SASL_LOG_ERR:
			fprintf(stderr, "error: ");
			break;
		case SASL_LOG_WARNING:
			fprintf(stderr, "warning: ");
			break;
		default:
			return SASL_OK; /* ignore comments */
	}
	fprintf(stderr, "%s\n", message);
	return SASL_OK;
}

int spdbm_putsecret(void *context, const char *mechanism,
	const char *auth_identity, const char *realm,
	const sasl_secret_t *secret) {
/* we ignore realm and auth_identity arguments */

	struct secret_str *my_secret = context;
	int len;

	if ((my_secret == NULL) || (secret == NULL))
		return SASL_FAIL;
	if ((secret->len + strlen(mechanism) +  sizeof(unsigned long int) +
		my_secret->len + 1) > sizeof(my_secret->secret))
		return SASL_BUFOVER;
	len = my_secret->len;
	strcpy(&my_secret->secret[len], mechanism);
	len += (strlen(mechanism) + 1);
	*((unsigned long int *)(&my_secret->secret[len])) = secret->len;
	len += sizeof(unsigned long int);
	memcpy(&my_secret->secret[len], secret->data, secret->len);
	my_secret->len = (len + secret->len);
	return SASL_OK;
}

int spdbm_sasl_init(char *realm) {
	int ret;

	static sasl_callback_t callbacks[] = {
		{SASL_CB_SERVER_PUTSECRET, spdbm_putsecret, &spdbm_sasl_secret},
		{SASL_CB_SERVER_GETSECRET, spdbm_getsecret, NULL},
		{SASL_CB_LOG, spdbm_log, NULL},
		{SASL_CB_LIST_END, NULL, NULL}
	};

	if ((ret = sasl_server_init(callbacks, "spdbm")) != SASL_OK) {
		print_errors("sasl_server_init", ret);
		return -1;
	}
	if ((ret = sasl_server_new("spdbm", NULL, realm, NULL,
		SASL_SECURITY_LAYER, &connection)) != SASL_OK) {
		print_errors("sasl_server_new", ret);
		sasl_done();
		return -1;
	}
	return 0;
}

void spdbm_sasl_done(void) {
	sasl_dispose(&connection);
	sasl_done();
}

int spdbm_sasl_create_secret(char *username, char *password, char *sasl_output,
	int *sasl_length) {
	int ret;

	spdbm_sasl_secret.len = 0;
	if ((ret = sasl_setpass(connection, username, password,
		strlen(password), 0, NULL)) != SASL_OK) {
		memset(&spdbm_sasl_secret, 0, sizeof(spdbm_sasl_secret));
		print_errors("sasl_setpass", ret);
		return -1;
	}
	if (spdbm_sasl_secret.len > SASLSECRETLN) {
		memset(&spdbm_sasl_secret, 0, sizeof(spdbm_sasl_secret));
		fprintf(stderr, "weird error: contact author of program\n");
		return -1;
	}
	memcpy(sasl_output, spdbm_sasl_secret.secret, spdbm_sasl_secret.len);
	*sasl_length = spdbm_sasl_secret.len;
	memset(&spdbm_sasl_secret, 0, sizeof(spdbm_sasl_secret));
	return 0;
}
# endif /* SPSASL */

int dpuid_addentry(MYDBM *db, int replace, char *username, char *password,
		    char *apop_secret, char *sasl_secret, char *maildrop_name,
		    char *maildrop_type, int encrypted) {
	char buf[MAXARGLN + PASSWORDLN + SASLSECRETLN + PATH_MAX +
		MAXMDTYPENAMELENGTH + (sizeof(unsigned long int) * 5) + 1];
	char *act;
	char value[MAXARGLN +PASSWORDLN + SASLSECRETLN + PATH_MAX + 
		MAXMDTYPENAMELENGTH + (sizeof(unsigned long int) * 5) + 1];
	int valuesize = sizeof(value) - 1, result, nr, sasl_done = 0;
	unsigned long int fieldln;
	/* password+apop_secret+sasl_secret+maildrop_name+maildrop_type */
	char salt[13];
# ifdef SPSASL
	char sasl_output[SASLSECRETLN];
	int sasl_length;
# endif
	struct {
		char *buf;
		int bufsize;
	} fields[5] = {
		{password, PASSWORDLN - 1},
		{apop_secret, MAXARGLN},
		{sasl_secret, SASLSECRETLN},
		{maildrop_name, PATH_MAX - 1},
		{maildrop_type, MAXMDTYPENAMELENGTH - 1}
	};

	result = 0;
	for (act = apop_secret; *act != 0; act++) {
		*act ^= 0xff;
		result++;
	} /* result = strlen(apop_secret) */
# ifndef SPSASL
	sasl_secret[0] = 0;
# endif
	nr = 0;
	if (!replace)
		if (mydbm_fetch(db, username, value, &valuesize) >= 0) {
			act = value;
			while (valuesize >= sizeof(unsigned long int)) {
				fieldln = *((unsigned long int *) act);
				*act = 0; /* end string with NUL */
				act += sizeof(unsigned long int);
				valuesize -= sizeof(unsigned long int);
				if ((fieldln > valuesize) || 
				    (fieldln > fields[nr].bufsize)) {
					fprintf(stderr, "database damaged\n");
					memset(value, 0, sizeof(value));
					return -1;
				}
				if (fields[nr].buf[0] == 0) {
					fields[nr].bufsize = fieldln;
					fields[nr].buf = act;
					if (nr == 0) /* password */
						encrypted = 1;
					if (nr == 1) /* APOP */
						result = fieldln;
					if (nr == 2) /* SASL */
						sasl_done = 1;
				}
				act += fieldln;
				valuesize -= fieldln;
				nr++;
			}
			if ((valuesize != 0) || (nr != 5)) {
				fprintf(stderr, "database damaged\n");
				memset(value, 0, sizeof(value));
				return -1;
			}
			*act = 0; /* end maildrop_type with NUL */
		}
	valuesize = 0;
	if ((encrypted) || (fields[0].buf[0] == 0))
		store_field(buf, fields[0].buf, strlen(fields[0].buf),
			&valuesize);
	else {
		act = md5_crypt(fields[0].buf, generate_salt(salt));
		store_field(buf, act, strlen(act), &valuesize);
		memset(salt, 0, sizeof(salt));
		memset(act, 0, strlen(act));
	}
	store_field(buf, fields[1].buf, result, &valuesize);
	/* always true if SPSASL is not defined */
	if (((!sasl_done) && (fields[2].buf[0] == 0)) || (sasl_done))
		store_field(buf, fields[2].buf, sasl_done ? fields[2].bufsize :
		0, &valuesize);
# ifdef SPSASL
	else { /* secret generation */
		result = spdbm_sasl_create_secret(username, fields[2].buf,
			sasl_output, &sasl_length);
		if (result == 0)
			store_field(buf, sasl_output, sasl_length, &valuesize);
		memset(sasl_output, 0, sizeof(sasl_output));
		if (result < 0) {
			memset(value, 0, sizeof(value));
			memset(buf, 0, sizeof(buf));
			return -1;
		}
	}
# endif	
	store_field(buf, fields[3].buf, strlen(fields[3].buf), &valuesize);
	store_field(buf, fields[4].buf, strlen(fields[4].buf), &valuesize);
	memset(value, 0, sizeof(value));
	result = mydbm_store(db, username, buf, valuesize, 1);
	memset(buf, 0, sizeof(buf));
	return (result < 0) ? -1 : 0;
}

void dpuid_manage(char *dbfile, int create, int replace,
		  char *username, char *password, char *apop_secret,
		  char *sasl_secret, char *maildrop_name, char *maildrop_type,
		  int encrypted) {
	MYDBM db;
	char buf[3 * (MAXARGLN + 1) + PATH_MAX + 1 + MAXMDTYPENAMELENGTH + 1];
	/* password:apop_secret:sasl_secret:maildrop_name:maildrop_type\0 */
	int result;
	char tmppassword[MAXARGLN + 1] = {0, };
	char tmpapop_secret[MAXARGLN + 1] = {0, };
	char tmpsasl_secret[MAXARGLN + 1] = {0, };
	char tmpusername[MAXARGLN + 1];
	char tmpmaildrop_name[PATH_MAX];
	char tmpmaildrop_type[MAXMDTYPENAMELENGTH];
	char *tmp, *tmp2;

	if (username[0] == 0) {
		memset(apop_secret, 0, strlen(apop_secret));
		memset(sasl_secret, 0, strlen(sasl_secret));
		memset(password, 0, strlen(password));
	}
	if (mydbm_open(&db, dbfile, create) < 0)
		exit(1);
	if (username[0] != 0) {
		if (password[0] == 0) {
			tmp = getpass("Enter NEW password: ");
			strncpy(tmppassword, tmp, sizeof(tmppassword) - 1);
			tmp = getpass("Reenter NEW password: ");
			encrypted = 0;
			result = strcmp(tmppassword, tmp);
			memset(tmp, 0, strlen(tmp));
			if (result != 0) {
				memset(tmppassword, 0, sizeof(tmppassword));
				mydbm_close(&db);
				fprintf(stderr, "Passwords don't match\n");
				exit(1);
			}
		}
# ifdef APOP
		if (apop_secret[0] == 0) {
			tmp = getpass("Enter NEW APOP secret: ");
			strncpy(tmpapop_secret, tmp,
				sizeof(tmpapop_secret) - 1);
			tmp = getpass("Reenter NEW APOP secret: ");
			result = strcmp(tmpapop_secret, tmp);
			memset(tmp, 0, strlen(tmp));
			if (result != 0) {
				memset(tmppassword, 0, sizeof(tmppassword));
				memset(tmpsasl_secret, 0, sizeof(tmpsasl_secret));
				memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
				mydbm_close(&db);
				fprintf(stderr, "APOP secrets don't match\n");
				exit(1);
			}
		}
# endif
# ifdef SPSASL
		if (sasl_secret[0] == 0) {
			tmp = getpass("Enter NEW SASL secret: ");
			strncpy(tmpsasl_secret, tmp,
				sizeof(tmpsasl_secret) - 1);
			tmp = getpass("Reenter NEW SASL secret: ");
			result = strcmp(tmpsasl_secret, tmp);
			memset(tmp, 0, strlen(tmp));
			if (result != 0) {
				memset(tmppassword, 0, sizeof(tmppassword));
				memset(tmpsasl_secret, 0,
					sizeof(tmpsasl_secret));
				memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
				mydbm_close(&db);
				fprintf(stderr, "SASL secrets don't match\n");
				exit(1);
			}
		}
# endif
		result = dpuid_addentry(&db, replace, username,
			(password[0] == 0) ? tmppassword : password,
			(apop_secret[0] == 0) ? tmpapop_secret : apop_secret,
			(sasl_secret[0] == 0) ? tmpsasl_secret : sasl_secret, 
			maildrop_name, maildrop_type, encrypted);
		memset(apop_secret, 0, strlen(apop_secret));
		memset(password, 0, strlen(password));
		memset(sasl_secret, 0, strlen(sasl_secret));
		memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
		memset(tmppassword, 0, sizeof(tmppassword));
		memset(tmpsasl_secret, 0, sizeof(tmpsasl_secret));
		mydbm_close(&db);
		if (result == 0)
			fprintf(stderr, "added/modified/replaced entry\n");
		else {
			fprintf(stderr, "error\n");
			exit(1);
		}
		return;
	}
	
	while (fgets(buf, sizeof(buf), stdin) != NULL) {
		if (strlen(buf) == (sizeof(buf) - 1)) {
			fprintf(stderr, "line too long\n");
			mydbm_close(&db);
			exit(1);
		}
		tmp2 = buf + strlen(buf) - 1;
		if (*tmp2 == '\n')
			*tmp2 = 0;
		if ((tmp2 = strchr(buf, ':')) == NULL) {
			fprintf(stderr, "error in entry\n");
			exit(1);
		}
		*tmp2 = 0;
		if (strlen(buf) >= sizeof(tmpusername)) {
			fprintf(stderr, "username too long\n");
			exit(1);
		}
		strcpy(tmpusername, buf);
		tmp = tmp2 + 1;
		
		
		if ((tmp2 = strchr(tmp, ':')) == NULL) {
			fprintf(stderr, "error in entry\n");
			exit(1);
		}
		*tmp2 = 0;
		if (strlen(tmp) >= sizeof(tmppassword)) {
			fprintf(stderr, "password too long\n");
			exit(1);
		}
		strcpy(tmppassword, tmp);
		tmp = tmp2 + 1;
		
		
		if ((tmp2 = strchr(tmp, ':')) == NULL) {
			fprintf(stderr, "error in entry\n");
			exit(1);
		}
		*tmp2 = 0;
		if (strlen(tmp) >= sizeof(tmpapop_secret)) {
			fprintf(stderr, "APOP secret too long\n");
			exit(1);
		}
		strcpy(tmpapop_secret, tmp);
		tmp = tmp2 + 1;
		
		if ((tmp2 = strchr(tmp, ':')) == NULL) {
			fprintf(stderr, "error in entry\n");
			exit(1);
		}
		*tmp2 = 0;
		if (strlen(tmp) >= sizeof(tmpsasl_secret)) {
			fprintf(stderr, "SASL secret too long\n");
			exit(1);
		}
		strcpy(tmpsasl_secret, tmp);
		tmp = tmp2 + 1;
		
		if ((tmp2 = strchr(tmp, ':')) == NULL) {
			fprintf(stderr, "error in entry\n");
			exit(1);
		}
		*tmp2 = 0;
		if (strlen(tmp) >= sizeof(tmpmaildrop_name)) {
			fprintf(stderr, "maildrop name too long\n");
			exit(1);
		}
		strcpy(tmpmaildrop_name, tmp);
		tmp = tmp2 + 1;
		
		
		if (strlen(tmp) >= sizeof(tmpmaildrop_type)) {
			fprintf(stderr, "maildrop type too long\n");
			exit(1);
		}
		strcpy(tmpmaildrop_type, tmp);
		result = dpuid_addentry(&db, replace, tmpusername, tmppassword,
			tmpapop_secret, tmpsasl_secret, tmpmaildrop_name,
			tmpmaildrop_type, encrypted);
		memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
		memset(tmppassword, 0, sizeof(tmppassword));
		memset(tmpsasl_secret, 0, sizeof(tmpsasl_secret));
		if (result < 0) {
			fprintf(stderr, "error\n");
			exit(1);
		}
	}
	mydbm_close(&db);
}
#endif /* DPUID */

int main(int argc, char **argv) {
	int c, replace = 0, create = DB_WRITER, mode = 0;
	char dbfile[PATH_MAX] = {0};
	char username[MAXARGLN + 1] = {0};
#ifdef MAPPING
	char mapname[MAXARGLN + 1] = {0};
#endif
#ifdef DPUID
	char maildrop_name[PATH_MAX] = {0};
	char maildrop_type[MAXMDTYPENAMELENGTH] = {0};
	char password[MAXARGLN + 1] = {0};
	int encrypted = 0;
	char apop_secret[MAXARGLN + 1] = {0};
	char sasl_secret[MAXARGLN + 1] = {0};
# ifdef SPSASL
	char *realm = NULL;
# endif
# ifdef MAPPING
	int dbtype = 0; /* mapping database */
# endif /* MAPPING */
#endif /* DPUID */
	char options[128]; /* enough long */

	fprintf(stderr,
"Solid POP3 database management utility\n\n");
	strcpy(options, "f:rcodl");
#ifdef MAPPING
	strcat(options, "m:");
#endif

#ifdef DPUID
	strcat(options, "p:n:t:e");
# ifdef APOP
	strcat(options, "a:");
# endif
# ifdef SPSASL
	strcat(options, "s:b:");
# endif
# ifdef MAPPING
	strcat(options, "u");
# endif
#endif

	while ((c = getopt(argc, argv, options)) != EOF)
		switch (c) {
			case 'f':
				if (strlen(optarg) >= sizeof(dbfile)) {
					fprintf(stderr, "filename too long\n");
					exit(1);
				}
				strcpy(dbfile, optarg);
				break;
#ifdef MAPPING
			case 'm':
				if (strlen(optarg) >= sizeof(mapname)) {
					fprintf(stderr, "mapname too long\n");
					exit(1);
				}
				strcpy(mapname, optarg);
				break;
#endif
#ifdef DPUID
			case 'p':
				if (strlen(optarg) >= sizeof(password)) {
					fprintf(stderr, "password too long\n");
					exit(1);
				}
				strcpy(password, optarg);
				break;
			case 'n':
				if (strlen(optarg) >= sizeof(maildrop_name)) {
					fprintf(stderr, "maildrop name too "
						"long\n");
					exit(1);
				}
				strcpy(maildrop_name, optarg);
				break;
			case 't':
				if (strlen(optarg) >= sizeof(maildrop_type)) {
					fprintf(stderr, "maildrop type too "
						"long\n");
					exit(1);
				}
				strcpy(maildrop_type, optarg);
				break;
			case 'e':
				encrypted = 1;
				break;
# ifdef APOP
			case 'a':
				if (strlen(optarg) >= sizeof(apop_secret)) {
					fprintf(stderr, "APOP secret too "
						"long\n");
					exit(1);
				}
				strcpy(apop_secret, optarg);
				break;
# endif /* APOP */
# ifdef SPSASL
			case 's':
				if (strlen(optarg) >= sizeof(sasl_secret)) {
					fprintf(stderr, "SASL secret too "
					"long\n");
					exit(1);
				}
				strcpy(sasl_secret, optarg);
				break;
			case 'b':
				realm = optarg;
				break;
# endif
# ifdef MAPPING
			case 'u':
				dbtype = 1;
				break;
# endif /* MAPPING */
#endif /* DPUID */
			case 'r':
				replace = 1;
				break;
			case 'c':
				create = DB_CREAT;
				break;
			case 'o':
				create = DB_NEWDB;
				break;
			case 'd':
				mode = 1;
				break;
			case 'l':
				mode = 2;
				break;
			case '?':
				usage(argv[0]);
				exit(1);
			default:
				fprintf(stderr, "unexpected error - "
					"contact author of program\n");
				exit(1);
		} /* switch */
	if (dbfile[0] == 0) {
		fprintf(stderr, "command line error: database filename not "
			"specified\n");
		usage(argv[0]);
		exit(1);
	}
	if (optind < argc) {
		if (strlen(argv[optind]) >= sizeof(username)) {
			fprintf(stderr, "username too long\n");
			exit(1);
		}
		strcpy(username, argv[optind]);
		optind++;
	}
	if (optind != argc) {
		fprintf(stderr, "command line error: too many arguments\n");
		usage(argv[0]);
		exit(1);
	}
	if (mode == 2) {
		list_entries(dbfile);
		exit(0);
	}
	if (mode == 1) {
		delete_entries(dbfile, (username[0] == 0) ? NULL : username);
		exit(0);
	}
#ifdef SPSASL
	if (spdbm_sasl_init(realm) < 0)
		exit(1);
	if (atexit(spdbm_sasl_done) < 0) {
		perror("atexit");
		spdbm_sasl_done();
		exit(1);
	}
#endif
#if defined(MAPPING) && defined(DPUID)
	if (dbtype == 1)
#endif
#ifdef DPUID
	{
		dpuid_manage(dbfile, create, replace,
			     username, password, apop_secret, sasl_secret, 
			     maildrop_name, maildrop_type, encrypted);
		exit(0);
	}
#endif
#ifdef MAPPING
	if ((username[0] != 0) && (mapname[0] == 0)) {
		fprintf(stderr, "command line error: mapname not specified\n");
		usage(argv[0]);
		exit(1);
	}
	if ((mapname[0] != 0) && (username[0] == 0)) {
		fprintf(stderr, "command line error: username not specified\n");
		usage(argv[0]);
		exit(1);
	}
	if (username[0] == 0)
		map_add_entries(dbfile, replace, create, NULL, NULL);
	else
		map_add_entries(dbfile, replace, create, username, mapname);
	exit(0);
#endif /* MAPPING */
}
