static const char rcsid[] = "$Id: spgetpwnam.c,v 1.7 2001/04/22 08:57:10 siefca Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include "log.h"
#include "configfile.h"

#include <pwd.h>
#ifdef AUTHONLY
#include <grp.h>
#endif

#ifdef MD5PASSWORD
# include "md5_crypt.h"

char sppassword[PASSWORDLN];
#endif
#ifdef AUTHONLY
extern int authonly;
extern struct passwd authonly_pw;
#endif
#ifdef DPUID
char dpuid_user[MAXARGLN + 1];
int dpuid_enabled = 0;
char dpuid_db[PATH_MAX];
char dpuid_log[MAXARGLN + 2];
extern char maildrop_name[];
extern char maildrop_type[];
# ifdef APOP
extern char apop_secret[];
# endif
# ifdef SPSASL
extern char sasl_secret[];
extern int sasl_secret_len;
# endif
# ifdef USERCONFIG
extern int useroverride;
# endif
# include "mydbm.h"
#endif

struct passwd pwentry;

struct passwd *spgetpwnam(const char *name, int *rootpriv) {
/* rootpriv - do we need root privileges to authenticate user?
   we don't when dpuid is used*/
	struct passwd *tmpentry;
	struct group *tmpgrentry;

#ifdef AUTHONLY
	if (authonly) {
	    bzero ((void*)&authonly_pw, sizeof(authonly_pw));
	    if ((tmpentry = getpwnam(authonly_user)) == NULL) {
		    pop_log(pop_priority, "user %s not found", authonly_user);
		    pop_error("getpwnam");
		    exit(1);
		    }
	    authonly_pw.pw_uid = tmpentry->pw_uid;
	    
	    if (authonly_fakehome == NULL)
		    authonly_pw.pw_dir = tmpentry->pw_dir;
	    else
		    authonly_pw.pw_dir = authonly_fakehome;

	    if (authonly_group == NULL)
		    authonly_pw.pw_gid = tmpentry->pw_gid;
	    else
		    {
		    if ((tmpgrentry = getgrnam(authonly_group)) == NULL) {
			 pop_log(pop_priority, "group %s not found",
				authonly_group);
			pop_error("getgrnam");
			exit(1);
			}
		    authonly_pw.pw_gid = tmpgrentry->gr_gid;
		    }
	    authonly_pw.pw_gecos = authonly_pw.pw_name 
		                 = authonly_pw.pw_passwd = NULL;
				 
	    return &authonly_pw;
	    }
#endif
#ifdef DPUID

	if (dpuid_enabled) {
		if (dpuid_db[0] == 0) {
			pop_log(pop_priority, "spgetpwnam: dpuid_dbname "
				"not specified");
			return NULL;
		}
		if (dpuid_user[0] == 0) {
		        pop_log(pop_priority, "spgetpwnam: dpuid_user not "
				"specified");
			return NULL;
		}
		if ((tmpentry = getpwnam(dpuid_user)) == NULL) {
			pop_log(pop_priority, "spgetpwnam: can't find user: "
				"%.40s", dpuid_user);
    			return NULL;
		}
		pwentry.pw_dir = tmpentry->pw_dir;
		pwentry.pw_shell = pwentry.pw_gecos = pwentry.pw_name = NULL;
		pwentry.pw_uid = tmpentry->pw_uid;
		pwentry.pw_gid = tmpentry->pw_gid;
		pwentry.pw_passwd = NULL; /* it will be set just before 
		authentication, after dropping privileges */
		if (rootpriv != NULL)
			*rootpriv = 0;
		return &pwentry;
	}
#endif
	if (rootpriv != NULL)
		*rootpriv = 1;
	if ((tmpentry = getpwnam(name)) == NULL)
		pop_log(pop_priority, "spgetpwnam: can't find user: %.40s",
			name);
	return tmpentry;
}

#ifdef DPUID
void dpuid_db_damaged(char *buf, int size) {
	pop_log(pop_priority, "spgetpwnam: dpuid database is damaged");
	memset(buf, 0, size);
	return;
}

int dpuid_set(char *name) {
	char tmpbuf[sizeof(sppassword) + MAXARGLN + SASLSECRETLN + PATH_MAX + 
		MAXMDTYPENAMELENGTH + (5 * sizeof(unsigned long int)) + 1];
	/* passwd,apopsecret,saslsecret,maildrop_name,maildrop_type */
	int valuesize = sizeof(tmpbuf) - 1, nr = 0;
	unsigned long int length;
	MYDBM db;
	char *act = tmpbuf;
	struct {
		char *buf;
		int bufsize;
	} fields[] = {
		{NULL, sizeof(sppassword) - 1},
		{NULL, MAXARGLN},
		{NULL, SASLSECRETLN},
		{NULL, PATH_MAX - 1},
		{NULL, MAXMDTYPENAMELENGTH - 1}
	};

	dpuid_log[0] = 0;
	strcat(dpuid_log, "/");
	strncat(dpuid_log, dpuid_user, sizeof(dpuid_log) - 2);

# ifdef USERCONFIG
	useroverride = 0; /* when we use dpuid domain maintainer shouldn't be
			     able to mess anything - APOP secret for example */
# endif
	if (mydbm_open(&db, dpuid_db, DB_READER) < 0)
		return -1;
	if (mydbm_fetch(&db, name, tmpbuf, &valuesize) < 0) {
		memset(tmpbuf, 0, sizeof(tmpbuf));
		return -1;
	}
	mydbm_close(&db);
	while (valuesize >= sizeof(unsigned long int)) {
		length = *((unsigned long int *) act);
		*act = 0; /* end string with NUL */
		act += sizeof(unsigned long int);
		valuesize -= sizeof(unsigned long int);
		if ((length > valuesize) || (length > fields[nr].bufsize)) {
			dpuid_db_damaged(tmpbuf, sizeof(tmpbuf));
			return -1;
		}
		fields[nr].bufsize = length;
		fields[nr].buf = act;
		nr++;
		act += length;
		valuesize -= length;
	}
	*act = 0;
	if ((valuesize != 0) || (nr != 5)) {
		dpuid_db_damaged(tmpbuf, sizeof(tmpbuf));
		return -1;
	}
	if (fields[4].buf[0]) {
		if (check_maildrop_type(fields[4].buf) < 0) {
			pop_log(pop_priority,  "spgetpwnam: unknown maildrop "
				"type: %.40s", fields[4].buf);
			memset(tmpbuf, 0, sizeof(tmpbuf));
			return -1;
		}
		strcpy(maildrop_type, fields[4].buf);
	}
	if (fields[3].buf[0])
		strcpy(maildrop_name, fields[3].buf);
# ifdef APOP
	for (act = fields[1].buf; fields[1].bufsize > 0; fields[1].bufsize--)
			*act ^= 0xff;
	strcpy(apop_secret, fields[1].buf);
# endif
# ifdef SPSASL
	memcpy(sasl_secret, fields[2].buf, fields[2].bufsize);
	sasl_secret_len = fields[2].bufsize;
# endif
	strncpy(sppassword, fields[0].buf, sizeof(sppassword));
	/* pad with NULs */
	memset(tmpbuf, 0, sizeof(tmpbuf));
	return 0;
}
#endif /* DPUID */

#ifdef MD5PASSWORD
int verify_password(char *password) { /* verify against sppassword */
	char *tmp = md5_crypt(password, sppassword);
	int result;
	
	memset(password, 0, strlen(password));
	if (tmp != NULL)
		result = strcmp(tmp, sppassword);
	else
		result = -1;
	memset(sppassword, 0, sizeof(sppassword));
	if (tmp != NULL)
		memset(tmp, 0, 3 + 9 + 24 + 2);
	return ((result == 0) ? 0 : -1);
}
#endif
