static const char rcsid[] = "$Id: sptls.c,v 1.12 2000/09/13 21:03:17 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include "log.h"
#include "response.h"
#include "sktbuff.h"
#include "sptls.h"
#ifdef SPSASL
# include "spsasl.h"
#endif

#include <fcntl.h>

#include <openssl/lhash.h>
#include <openssl/bn.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/ssl.h>
#include <openssl/rand.h>
#include <openssl/dh.h>

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

extern int encenabled;
extern void check_wccount(void);

int _tls_pending = 1;
int tls_loglevel = 0;
int allowtls = 1, requiretls = 0, allowtmprsa = 1;
int verifydepth = 1;
int askcert = 0, requirecert = 0;
char certfile[PATH_MAX], keyfile[PATH_MAX];
char dcertfile[PATH_MAX], dkeyfile[PATH_MAX];
char cafile[PATH_MAX], capath[PATH_MAX];
char dh512paramfile[PATH_MAX], dh1024paramfile[PATH_MAX];
char cipherlist[PATH_MAX];
SSL_CTX *tls_ctx = NULL;
SSL *tls_connection = NULL;
DH *dh512 = NULL, *dh1024 = NULL;
char rand_source[PATH_MAX];
int rand_bytes = 255;

static unsigned char dh512_p[] = {
	0xF5,0x2A,0xFF,0x3C,0xE1,0xB1,0x29,0x40,0x18,0x11,0x8D,0x7C,
	0x84,0xA7,0x0A,0x72,0xD6,0x86,0xC4,0x03,0x19,0xC8,0x07,0x29,
	0x7A,0xCA,0x95,0x0C,0xD9,0x96,0x9F,0xAB,0xD0,0x0A,0x50,0x9B,
	0x02,0x46,0xD3,0x08,0x3D,0x66,0xA4,0x5D,0x41,0x9F,0x9C,0x7C,
	0xBD,0x89,0x4B,0x22,0x19,0x26,0xBA,0xAB,0xA2,0x5E,0xC3,0x55,
	0xE9,0x2A,0x05,0x5F,
};
static unsigned char dh512_g[] = {
	0x02,
};
static unsigned char dh1024_p[] = {
	0xF4,0x88,0xFD,0x58,0x4E,0x49,0xDB,0xCD,0x20,0xB4,0x9D,0xE4,
	0x91,0x07,0x36,0x6B,0x33,0x6C,0x38,0x0D,0x45,0x1D,0x0F,0x7C,
	0x88,0xB3,0x1C,0x7C,0x5B,0x2D,0x8E,0xF6,0xF3,0xC9,0x23,0xC0,
	0x43,0xF0,0xA5,0x5B,0x18,0x8D,0x8E,0xBB,0x55,0x8C,0xB8,0x5D,
	0x38,0xD3,0x34,0xFD,0x7C,0x17,0x57,0x43,0xA3,0x1D,0x18,0x6C,
	0xDE,0x33,0x21,0x2C,0xB5,0x2A,0xFF,0x3C,0xE1,0xB1,0x29,0x40,
	0x18,0x11,0x8D,0x7C,0x84,0xA7,0x0A,0x72,0xD6,0x86,0xC4,0x03,
	0x19,0xC8,0x07,0x29,0x7A,0xCA,0x95,0x0C,0xD9,0x96,0x9F,0xAB,
	0xD0,0x0A,0x50,0x9B,0x02,0x46,0xD3,0x08,0x3D,0x66,0xA4,0x5D,
	0x41,0x9F,0x9C,0x7C,0xBD,0x89,0x4B,0x22,0x19,0x26,0xBA,0xAB,
	0xA2,0x5E,0xC3,0x55,0xE9,0x2F,0x78,0xC7,
};
static unsigned char dh1024_g[] = {
	0x02,
};
DH *get_dh512(void) {
	DH *dh;

	if ((dh = DH_new()) == NULL)
		return(NULL);
	dh->p = BN_bin2bn(dh512_p, sizeof(dh512_p), NULL);
	dh->g = BN_bin2bn(dh512_g, sizeof(dh512_g), NULL);
	if ((dh->p == NULL) || (dh->g == NULL))
		return(NULL);
	return(dh);
}

DH *get_dh1024(void) {
	DH *dh;

	if ((dh = DH_new()) == NULL)
		return(NULL);
	dh->p = BN_bin2bn(dh1024_p, sizeof(dh1024_p), NULL);
	dh->g = BN_bin2bn(dh1024_g, sizeof(dh1024_g), NULL);
	if ((dh->p == NULL) || (dh->g == NULL))
		return(NULL);
	return(dh);
}

void tls_print_errors(void) {
	unsigned long l;
	const char *file;
	const char *data;
	int line;
	int flags;
	unsigned long es;
	
	es = CRYPTO_thread_id();
	while ((l = ERR_get_error_line_data(&file, &line, &data, &flags)) != 0) {
		if (flags & ERR_TXT_STRING)
			pop_log(pop_priority,
				"sptls: %lu:%.256s:%.256s:%d:%.256s",
				es, ERR_error_string(l, NULL), file, line,
				data);
		else
			pop_log(pop_priority, "sptls: %lu:%.256s:%.256s:%d",
				es, ERR_error_string(l, NULL), file, line);
	}
}

void tls_rand_seed(void) {
	struct {
		struct timeval tv;
		pid_t pid;
		pid_t ppid;
	} randseed;
	int randbytes;

#ifdef GETTIMEOFDAY_TZ
	gettimeofday(&randseed.tv, NULL);
#else
	gettimeofday(&randseed.tv);
#endif
	randseed.pid = getpid();
	randseed.ppid = getppid();
	RAND_add(&randseed, sizeof(randseed), 0.5); /* 4 bits of entropy
	should be secure */
	if (rand_source[0]) {
		if ((randbytes = RAND_egd(rand_source)) < 0)
			randbytes = RAND_load_file(rand_source, rand_bytes);
		if (randbytes < 0) {
			pop_log(pop_priority, "sptls: can't seed PRNG");
			tls_print_errors();
		} else
			pop_log(pop_priority,
				"sptls: seeding PRNG with %d external bytes",
				randbytes);
	}
}


DH *tls_tmp_dh_callback(SSL *s, int export, int keylength) {
		if (keylength <= 512)
			return dh512;
		else
			return dh1024;
}

int tls_verify_callback(int ok, X509_STORE_CTX *x509_ctx) {
	X509 *err_cert;
	char err_buf[256];
	int err, depth;
	
	err_cert = X509_STORE_CTX_get_current_cert(x509_ctx);
	err = X509_STORE_CTX_get_error(x509_ctx);
	depth = X509_STORE_CTX_get_error_depth(x509_ctx);
	X509_NAME_oneline(X509_get_subject_name(err_cert), err_buf,
		sizeof(err_buf));
	if (tls_loglevel >= 1)
		pop_log(pop_priority, "sptls: peer cert verify depth: %d: "
			"%.256s", depth, err_buf);
	if (!ok)
		pop_log(pop_priority, "sptls: verify error: num: %d: %.512s",
			err, X509_verify_cert_error_string(err));
	if (verifydepth < depth) {
		ok = 0;
		pop_log(pop_priority, "sptls: cert chain too long: %d", depth);
	}
	switch (x509_ctx->error) {
		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
			X509_NAME_oneline(X509_get_issuer_name(
				x509_ctx->current_cert), err_buf,
				sizeof(err_buf));
			pop_log(pop_priority, "sptls: issuer: %.256s", err_buf);
			break;
		case X509_V_ERR_CERT_NOT_YET_VALID:
		case X509_V_ERR_ERROR_IN_CERT_NOT_BEFORE_FIELD:
			pop_log(pop_priority, "sptls: cert not yet valid");
			break;
		case X509_V_ERR_CERT_HAS_EXPIRED:
		case X509_V_ERR_ERROR_IN_CERT_NOT_AFTER_FIELD:
			pop_log(pop_priority, "sptls: cert has expired");
			break;
	}
	if (tls_loglevel >= 1)
		pop_log(pop_priority, "sptls: verify return: %d", ok);
	return ok;
}

void tls_info_callback(SSL *s, int where, int ret) {
	char *str;
	int w = (where & (~SSL_ST_MASK));
	
	if (w & SSL_ST_CONNECT)
		str = "SSL_connect";
	else if (w & SSL_ST_ACCEPT)
		str = "SSL_accept";
	else
		str = "undefined";
	if (where & SSL_CB_LOOP) {
		if (tls_loglevel >= 2)
			pop_log(pop_priority, "sptls: %.20s: %.512s", str,
				SSL_state_string_long(s));
	} else if (where & SSL_CB_ALERT) {
		str = ((where & SSL_CB_READ) ? "read" : "write");
		if ((tls_loglevel >= 2) || ((ret & 0xff) !=
			SSL3_AD_CLOSE_NOTIFY))
			pop_log(pop_priority, "sptls: SSL3 alert: %.20s: %.512s"
				"%.512s", str, SSL_alert_type_string_long(ret),
				SSL_alert_desc_string_long(ret));
	} else if (where & SSL_CB_EXIT) {
		if (ret == 0)
			pop_log(pop_priority, "sptls: %.20s: failed in %.512s",
				str, SSL_state_string_long(s));
		else if (ret < 0)
			pop_log(pop_priority, "sptls: %.20s: error in %.512s",
				str, SSL_state_string_long(s));
	}
}

RSA *tls_tmp_rsa_callback(SSL *s, int export, int keylength) {
	static RSA *rsa_tmp = NULL;
	
	if (rsa_tmp == NULL)
		rsa_tmp = RSA_generate_key(keylength, RSA_F4, NULL, NULL);
	if (rsa_tmp != NULL)
		pop_log(pop_priority, "sptls: created %dbit temporary RSA key",
			keylength);
	else {
		pop_log(pop_priority, "sptls: can't create %dbit "
			"temporary RSA key", keylength);
		tls_print_errors();
	}
	return rsa_tmp;
}

int tls_no_passphrase_callback(char *buf, int num, int w, void *arg) {
	return -1;
}

int set_cert_stuff(char *acertfile, char *akeyfile) {
	if (acertfile[0]) {
		if (SSL_CTX_use_certificate_file(tls_ctx, acertfile,
			SSL_FILETYPE_PEM) <= 0) {
			pop_log(pop_priority, "sptls: unable to get "
				"certificate: %.1024s", acertfile);
			tls_print_errors();
			return -1;
		}
		if (akeyfile[0] == 0)
			strncat(akeyfile, acertfile, PATH_MAX - 1);
		if (SSL_CTX_use_PrivateKey_file(tls_ctx, akeyfile,
			SSL_FILETYPE_PEM) <= 0) {
			pop_log(pop_priority, "sptls: unable to get private "
				"key: %.1024s", akeyfile);
			tls_print_errors();
			return -1;
		}
		if (!SSL_CTX_check_private_key(tls_ctx)) {
			pop_log(pop_priority, "sptls: private key doesn't "
				"match the certificate public key");
			tls_print_errors();
			return -1;
		}
	}
	return 0;
}

void tls_done_serverengine(void) {
	if (tls_ctx != NULL)
		SSL_CTX_free(tls_ctx);
	tls_ctx = NULL;
}

int tls_init_serverengine(void) {
	if (tls_loglevel >= 2)
		pop_log(pop_priority, "sptls: starting TLS engine");
	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();
	if ((tls_ctx = SSL_CTX_new(SSLv23_server_method())) == NULL) {
		pop_log(pop_priority, "sptls: SSL_CTX_new failed");
		tls_print_errors();
		return -1;
	}
	SSL_CTX_set_options(tls_ctx, SSL_OP_ALL);
	SSL_CTX_set_info_callback(tls_ctx, tls_info_callback);
	SSL_CTX_sess_set_cache_size(tls_ctx, 1);
	SSL_CTX_set_default_passwd_cb(tls_ctx, tls_no_passphrase_callback);
	SSL_CTX_set_tmp_dh_callback(tls_ctx, tls_tmp_dh_callback);
	return 0;
}


void tls_stop_servertls(void) {
	if (tls_connection != NULL)
		SSL_free(tls_connection);
	tls_connection = NULL;
}

int tls_start_servertls(int *layerbits, char *authid) {
	/* strlen(authid) <= MAXARGLN */
	int sts;
	int algbits;
	SSL_CIPHER *cipher;
	X509 *peer;
	char tls_buf[256], tls_peer[256];
	int verifyflags = SSL_VERIFY_NONE;
	FILE *paramfile;
	
	tls_rand_seed();
	if (RAND_status() != 1) {
		pop_log(pop_priority, "sptls: insufficient random seeding");
		tls_print_errors();
		tls_done_serverengine();
		return -1;
	}
	if (cipherlist[0])
		if (SSL_CTX_set_cipher_list(tls_ctx, cipherlist) == 0) {
			tls_print_errors();
			tls_done_serverengine();
			return -1;
		}
	if (cafile[0] || capath[0])
		if (!SSL_CTX_load_verify_locations(tls_ctx,
			((cafile[0] == 0) ? NULL : cafile), ((capath[0] == 0) ?
			NULL : capath))) {
			pop_log(pop_priority, "sptls: can't load CA data");
			tls_print_errors();
			tls_done_serverengine();
			return -1;
		}
	if (set_cert_stuff(certfile, keyfile) < 0) {
		pop_log(pop_priority, "sptls: can't load RSA cert/key data");
		tls_done_serverengine();
		return -1;
	}
	if (set_cert_stuff(dcertfile, dkeyfile) < 0) {
		pop_log(pop_priority, "sptls: can't load DSA cert/key data");
		tls_done_serverengine();
		return -1;
	}
	if ((certfile[0] == 0) && (dcertfile[0] == 0)) {
		pop_log(pop_priority, "sptls: do need at least RSA or DSA "
			"cert/key data");
		tls_done_serverengine();
		return -1;
	}
	if (allowtmprsa)
		SSL_CTX_set_tmp_rsa_callback(tls_ctx, tls_tmp_rsa_callback);
	if (dh512paramfile[0]) {
		if ((paramfile = fopen(dh512paramfile, "r")) != NULL) {
			if ((dh512 = PEM_read_DHparams(paramfile, NULL, NULL,
			     NULL)) == NULL) {
				pop_log(pop_priority, 
				    "sptls: can't load 512bit DH parameters, "
				    "using built in parameters instead");
				tls_print_errors();
			}
			fclose(paramfile);
		} else {
			pop_log(pop_priority, 
			    "sptls: can't load 512bit DH parameters: %.1024s",
				dh512paramfile);
			pop_error("sptls: fopen");
		}
	}
	if (dh1024paramfile[0]) {
		if ((paramfile = fopen(dh1024paramfile, "r")) != NULL) {
			if ((dh1024 = PEM_read_DHparams(paramfile, NULL, NULL,
			     NULL)) == NULL) {
				pop_log(pop_priority, 
				    "sptls: can't load 1024bit DH parameters, "
				    "using built in parameters instead");
				tls_print_errors();
			}
			fclose(paramfile);
		} else {
			pop_log(pop_priority,
			    "sptls: can't load 1024bit DH parameters: %.1024s",
				dh1024paramfile);
			pop_error("sptls: fopen");
		}
	}
	if (dh512 == NULL)
		if ((dh512 = get_dh512()) == NULL) {
			pop_log(pop_priority,
			    "sptls: can't load built in 512bit DH parameters");
			tls_print_errors();
			tls_done_serverengine();
			return -1;
		}
	if (dh1024 == NULL)
		if ((dh1024 = get_dh1024()) == NULL) {
			pop_log(pop_priority,
			    "sptls: can't load built in 1024bit DH parameters");
			tls_print_errors();
			tls_done_serverengine();
			return -1;
		}
	if (askcert)
		verifyflags |= (SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE);
	if (requirecert)
		verifyflags |= (SSL_VERIFY_PEER | SSL_VERIFY_CLIENT_ONCE |
				SSL_VERIFY_FAIL_IF_NO_PEER_CERT);
	SSL_CTX_set_verify(tls_ctx, verifyflags, tls_verify_callback);
	SSL_CTX_set_client_CA_list(tls_ctx, SSL_load_client_CA_file(cafile[0]
		? cafile : NULL));
	if (tls_loglevel >= 1)
		pop_log(pop_priority, "sptls: setting up TLS connection");
	if ((tls_connection = SSL_new(tls_ctx)) == NULL) {
		pop_log(pop_priority, "sptls: SSL_new failed");
		tls_print_errors();
		tls_done_serverengine();
		return -1;
	}
	if (!SSL_set_rfd(tls_connection, 0)) {
		pop_log(pop_priority, "sptls: SSL_set_rfd failed");
		tls_print_errors();
		tls_stop_servertls();
		tls_done_serverengine();
		return -1;
	}
	if (!SSL_set_wfd(tls_connection, 1)) {
		pop_log(pop_priority, "sptls: SSL_set_wfd failed");
		tls_print_errors();
		tls_stop_servertls();
		tls_done_serverengine();
		return -1;
	}
	SSL_set_accept_state(tls_connection);
	if ((sts = SSL_accept(tls_connection)) <= 0) {
		pop_log(pop_priority, "sptls: SSL_accept error: %d", sts);
		tls_print_errors();
		tls_stop_servertls();
		tls_done_serverengine();
		return -1;
	}
	if ((peer = SSL_get_peer_certificate(tls_connection)) != NULL) {
		X509_NAME_oneline(X509_get_subject_name(peer), tls_buf,
			sizeof(tls_buf));
		if (tls_loglevel >= 2)
			pop_log(pop_priority, "sptls: subject: %.256s",
				tls_buf);
		X509_NAME_oneline(X509_get_issuer_name(peer), tls_buf,
			sizeof(tls_buf));
		if (tls_loglevel >= 2)
			pop_log(pop_priority, "sptls: issuer: %.256s",
				tls_buf);
		X509_NAME_get_text_by_NID(X509_get_subject_name(peer),
			NID_commonName, tls_peer, sizeof(tls_peer));
		X509_NAME_get_text_by_NID(X509_get_issuer_name(peer),
			NID_commonName, tls_buf, sizeof(tls_buf));
		if (tls_loglevel >= 3)
		pop_log(pop_priority, "sptls: subject_CN: %.256s, issuer_CN: "
			"%.256s", tls_peer, tls_buf);
		if (strlen(tls_peer) > MAXARGLN)
			pop_log(pop_priority, "sptls: issuer_CN is too long - "
				"can't use it to authentication");
		else
			strncpy(authid, tls_peer, MAXARGLN + 1);
			/* pad with NULs */
		X509_free(peer);
	}
	cipher = SSL_get_current_cipher(tls_connection);
	*layerbits = SSL_CIPHER_get_bits(cipher, &algbits);
	if (authid[0] == 0)
		pop_log(pop_priority, "sptls: TLS connection established: "
			"%.40s with cipher %.40s (%d/%d bits)",
			SSL_get_version(tls_connection),
			SSL_CIPHER_get_name(cipher), *layerbits, algbits);
	else
		pop_log(pop_priority, "sptls: TLS connection established: "
			"%.40s with cipher %.40s (%d/%d bits) authenticated "
			"as: %.40s", SSL_get_version(tls_connection),
			SSL_CIPHER_get_name(cipher), *layerbits, algbits,
			authid);
	return 0;
}

ssize_t tls_read(void *buf, size_t size, int block) {
	long flags;
	
	block = (block ? 1 : 0);
	if ((flags = fcntl(0, F_GETFL)) < 0) {
		pop_error("sptls: fcntl");
		return -1;
	}
	if (block)
		flags &= (~O_NONBLOCK);
	else
		flags |= O_NONBLOCK;
	if (fcntl(0, F_SETFL, flags) < 0) {
		pop_error("sptls: fcntl");
		return -1;
	}
	if ((flags = SSL_read(tls_connection, buf, size)) < 0) {
		if (block)
			return -1; /* error */
		switch(SSL_get_error(tls_connection, flags)) {
			case SSL_ERROR_WANT_READ:
				ERR_clear_error();
		    		_tls_pending = 0;
				return -2; /* don't block */
			case SSL_ERROR_WANT_WRITE:
				ERR_clear_error();
				return tls_read(buf, size, 1);
					/* retry in blocking mode */
			default:
				return -1; /* error */
		}
	}
	if (flags > 0)
		_tls_pending = 1;
	else
		_tls_pending = 0;
	return flags;
}

int tls_pending(void) {
	return _tls_pending;
}

ssize_t tls_write(void *buf, size_t size) {
	int result;
	long flags;
	
	if ((flags = fcntl(1, F_GETFL)) < 0) {
		pop_error("sptls: fcntl");
		return -1;
	}
	flags &= (~O_NONBLOCK);
	if (fcntl(1, F_SETFL, flags) < 0) {
		pop_error("sptls: fcntl");
		return -1;
	}
	result = SSL_write(tls_connection, buf, size);
	if (result < 0)
		return -1;
	else {
		return result;
	}
}

void tls_enable(int stls) {
	char authid[MAXARGLN + 1] = {0, };
	int lbits;

	if (tls_connection != NULL)
		return;
#ifndef STANDALONE
	if (tls_init_serverengine() < 0) {
		if (stls) {
		    send_error("error in TLS initialization");
		    if (buf_flush() < 0)
			    exit(1);
		    return;
		} else
		    exit(1);
	}
#endif
	if (stls) {
		send_ok("starting TLS");
		if (buf_flush() < 0)
			exit(1);
	}
	if (tls_start_servertls(&lbits, authid) < 0) {
		tls_done_serverengine();
		exit(1);
	}
	encenabled |= 2; /* TLS enabled */
#ifdef SPSASL
	set_secprops(((authid[0] == 0) ? NULL : authid), lbits);
#endif
}

void do_stls(char *arg) {
	if (tls_connection != NULL) {
		send_error("TLS layer is active");
		check_wccount();
		return;
	}
	if (allowtls || requiretls)
		tls_enable(1);
	else {
		send_error("STLS not allowed");
		check_wccount();
	}
}

