static const char rcsid[] = "$Id: userconfig.c,v 1.9 2000/09/29 10:31:54 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "maildrop.h"
#include "const.h"
#include "log.h"
#include "fdfgets.h"

extern char maildrop_name[];
extern char maildrop_type[];
#ifdef APOP
extern char apop_secret[];
#endif
#ifdef USERPASSWD
extern char sppassword[];
#endif
#ifdef SPSASL
extern char sasl_secret[];
extern int sasl_secret_len;
#endif

int parse_maildrop_cfg(unsigned int linenr, int *md_set,
		       char *tmpmaildrop_name, char *tmpmaildrop_type) {
	char *tmp;
	if (*md_set == 1) {
		pop_log(pop_priority, "user config: maildrop already set "
			"in user config file, line: %u", linenr);
		return -1;
	}
	if ((tmp = strtok(NULL, " \t")) == NULL) {
		pop_log(pop_priority, 
			"user config: argument missing in maildrop declaration "
			"in user config file, line: %u", linenr);
		return -1;
	}
	if (strlen(tmp) >= PATH_MAX) {
		pop_log(pop_priority, "user config: maildrop file name "
			"in user config file is too long, line: %u", linenr);
		return -1;
	}
	strcpy(tmpmaildrop_name, tmp);
	if ((tmp = strtok(NULL, " \t")) == NULL) {
		pop_log(pop_priority, "user config: argument missing "
			"in maildrop declaration in user config file, "
			"line: %u", linenr);
		return -1;
	}
	if (strlen(tmp) >= (MAXARGLN + 1)) {
		pop_log(pop_priority, "user config: maildrop type in user "
			"config file is too long, line: %u", linenr);
		return -1;
	}
#ifdef MDMAILBOX
	if (strcasecmp(tmp, "mailbox") != 0)
#endif
#ifdef MDMAILDIR
	if (strcasecmp(tmp, "maildir") != 0)
#endif
#ifdef MDCRLFMAILDIR
	if (strcasecmp(tmp, "crlfmaildir") != 0)
#endif
	{
		pop_log(pop_priority, "user config: no such maildrop type: "
			"%.40s, line: %u", tmp, linenr);
		return -1;
	}
	strcpy(tmpmaildrop_type, tmp);
	*md_set = 1;
	return 0;
}

#ifdef APOP
int parse_apop_cfg(unsigned int linenr, int *as_set, char *tmpapop_secret) {
	char *tmp;
	size_t tmp2;
	
	if (*as_set == 1) {
		pop_log(pop_priority, "user config: APOP secret already set "
			"in user config file, line: %u", linenr);
		return -1;
	}
	if ((tmp = strtok(NULL, " \t")) == NULL) {
		pop_log(pop_priority, "user config: argument missing, "
			"line: %u", linenr);
		return -1;
	}
	tmp2 = strlen(tmp);
	if (tmp2 & 1) {
		pop_log(pop_priority, "user config: length of encrypted APOP "
			"secret should be even number, line: %u", linenr);
		return -1;
	}
	if ((tmp2/2) >= (MAXARGLN + 1)) {
		pop_log(pop_priority, "user config: encrypted APOP secret "
			"is too long, line: %u", linenr);
		return -1;
	}
	for (tmp2 = 0; tmp2 < strlen(tmp); tmp2++)
		if (((tmp[tmp2] < 'a') || (tmp[tmp2] > 'z')) &&
		    (!isdigit(tmp[tmp2]))) {
			    pop_log(pop_priority, "user config: wrong character "
				    "in encrypted APOP secret, line: %u",
				    linenr);
			    return -1;
		    }
	memset(tmpapop_secret, 0, MAXARGLN + 1);
	for (tmp2 = 0; tmp2 < (strlen(tmp) / 2); tmp2++) {
		if (isdigit(tmp[tmp2 * 2]))
			tmpapop_secret[tmp2] = (tmp[tmp2 * 2] - '0') << 4;
		else
			tmpapop_secret[tmp2] = (tmp[tmp2 * 2] - 'a' + 10) << 4;
		if (isdigit(tmp[(tmp2 * 2) + 1]))
			tmpapop_secret[tmp2] |= (tmp[(tmp2 * 2) + 1] - '0');
		else
			tmpapop_secret[tmp2] |= (tmp[(tmp2 * 2) + 1] - 'a' + 10);
		tmpapop_secret[tmp2] ^= 0xff;
	}
	*as_set = 1;
	return 0;
}
#endif

#ifdef USERPASSWD
int parse_userpasswd_cfg(unsigned int linenr, int *pw_set,
	char *tmpsppassword) {
	char *tmp;
	
	if (*pw_set == 1) {
		pop_log(pop_priority, "user config: password already set "
			"in user config file, line: %u", linenr);
		return -1;
	}
	if ((tmp = strtok(NULL, " \t")) == NULL) {
		pop_log(pop_priority, "user config: argument missing, "
			"line: %u", linenr);
		return -1;
	}
	if (strlen(tmp) >= PASSWORDLN) {
		pop_log(pop_priority, "user config: password "
			"is too long, line: %u", linenr);
		return -1;
	}
	strncpy(tmpsppassword, tmp, PASSWORDLN); /* pad with NULs */
	*pw_set = 1;
	return 0;
}
#endif

void parse_user_cfg(char *homedir) {
	struct stat stbuf;
	char tmpmaildrop_name[PATH_MAX];
	char tmpmaildrop_type[MAXMDTYPENAMELENGTH];
#ifdef APOP
	char tmpapop_secret[MAXARGLN + 1];
	int as_set = 0;
#endif
#ifdef USERPASSWD
	char tmpsppassword[PASSWORDLN];
	int pw_set = 0;
#endif
	char cfgfile[PATH_MAX];
	char buf[128];
	int fd, md_set = 0;
	int linenr;
	ssize_t tmp;

#ifdef SPSASL
	void parse_user_sasl(void) {
		/* homedir has right mode */
		if ((strlen(homedir) + strlen(USERSASL) + 2) > 
			sizeof(cfgfile)) {
			pop_log(pop_priority, "user config: home directory "
				"name too long");
			return;
		}
		strcpy(cfgfile, homedir);
		strcat(cfgfile, "/");
		strcat(cfgfile, USERSASL);
		if ((fd = open(cfgfile, O_RDONLY)) < 0) {
			pop_log_dbg(pop_priority, "user config: can't open "
				"user SASL file");
			pop_error_dbg("user config: open");
			return;
		}
		if (fstat(fd, &stbuf) < 0) {
			close(fd);
			pop_log(pop_priority, "user config: can't stat user "
				"SASL file");
			pop_error("user config: fstat");
			return;
		}
		if (!S_ISREG(stbuf.st_mode)) {
			close(fd);
			pop_log(pop_priority, "user config: user SASL file "
				"is not regular file");
			return;
		}
		if (stbuf.st_mode & 0177) {
			close(fd);
			pop_log(pop_priority, "user config: user SASL file "
				"has wrong mode");
			return;
		}
		if (stbuf.st_size == 0) {
			close(fd);
			pop_log(pop_priority, "user config: user SASL file "
				"is empty");
			return;
		}
		if ((sasl_secret_len = read(fd, sasl_secret, SASLSECRETLN)) < 0) {
			sasl_secret_len = 0;
			close(fd);
			pop_log(pop_priority, "user config: can't read user "
				"SASL file");
			pop_error("read");
			return;
		}
		close(fd);
	}
#endif /* SPSASL */

	if (stat(homedir, &stbuf) < 0) {
		pop_log_dbg(pop_priority, "user config: can't stat user home "
			"directory: %.1024s", homedir);
		pop_error_dbg("user config: stat");
		return;
	}
	if (stbuf.st_mode & 2) {
		pop_log(pop_priority, "user config: user home directory is "
			"world writable");
		return;
	}
	if (stbuf.st_mode & 01000) {
		pop_log(pop_priority, "user config: user home directory has "
			"the sticky bit set");
		return;
	}
	if ((strlen(homedir) + strlen(USERCFG) + 2) > sizeof(cfgfile)) {
		pop_log(pop_priority, "user config: home directory name "
			"too long");
		return;
	}
#ifdef SPSASL
	parse_user_sasl();
#endif
	strcpy(cfgfile, homedir);
	strcat(cfgfile, "/");
	strcat(cfgfile, USERCFG);
	if ((fd = open(cfgfile, O_RDONLY)) < 0) {
		pop_log_dbg(pop_priority, "user config: can't open user "
			"config file");
		pop_error_dbg("user config: open");
		return;
	}
	if (fstat(fd, &stbuf) < 0) {
		close(fd);
		pop_log(pop_priority, "user config: can't stat user "
			"config file");
		pop_error("user config: fstat");
		return;
	}
	if (!S_ISREG(stbuf.st_mode)) {
		close(fd);
		pop_log(pop_priority, "user config: user config file is not "
			"regular file");
		return;
	}
	if (stbuf.st_mode & 0177) {
		close(fd);
		pop_log(pop_priority, "user config: user config file has "
			"wrong mode");
		return;
	}
	if (stbuf.st_size == 0) {
		close(fd);
		pop_log(pop_priority, "user config: user config file is empty");
		return;
	}
	fd_initfgets();
	linenr = 0;
	while (1) {
		linenr++;
		if ((tmp = fd_fgets(buf, sizeof(buf), fd, NULL)) < 0) {
			memset(buf, 0, sizeof(buf));
			close(fd);
			pop_log(pop_priority, "user config: can't read user "
				"config file");
			pop_error("user config: read");
#ifdef APOP
			memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
#endif
#ifdef USERPASSWD
			memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
			return;
		}
		if (tmp == 0)
			break;
		if ((tmp == sizeof(buf)) && (buf[sizeof(buf) - 1] != '\n')) {
			memset(buf, 0, sizeof(buf));
			close(fd);
			pop_log(pop_priority, "user config: line too long "
				"in user config file");
#ifdef APOP
			memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
#endif
#ifdef USERPASSWD
			memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
			return;
		}
		if ((tmp == 1) && (buf[0] == '\n'))
			continue;
		if (buf[tmp - 1] == '\n')
			buf[tmp - 1] = 0;
		else
			buf[tmp] = 0;
		strtok(buf, " \t");
		tmp = -2;
		if (strcasecmp(buf, "maildrop") == 0)
			tmp = parse_maildrop_cfg(linenr, &md_set,
				tmpmaildrop_name, tmpmaildrop_type);
#ifdef APOP
		if (strcasecmp(buf, "APOPsecret") == 0)
			tmp = parse_apop_cfg(linenr, &as_set, tmpapop_secret);
#endif
#ifdef USERPASSWD
		if (strcasecmp(buf, "Password") == 0)
			tmp = parse_userpasswd_cfg(linenr, &pw_set,
				tmpsppassword);
#endif
		memset(buf, 0, sizeof(buf));
		if (tmp == -2)
	    		pop_log(pop_priority, "user config: unknown option "
				"name, line: %u", linenr);
		if (tmp < 0) {
#ifdef APOP
			memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
#endif
#ifdef USERPASSWD
			memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif
			close(fd);
			return;
		}
	}
	close(fd);
	if (md_set == 1) {
		strcpy(maildrop_name, tmpmaildrop_name);
		strcpy(maildrop_type, tmpmaildrop_type);
	}
#ifdef APOP
	if (as_set == 1)
		strcpy(apop_secret, tmpapop_secret);
	memset(tmpapop_secret, 0, sizeof(tmpapop_secret));
#endif
#ifdef USERPASSWD
	if (pw_set == 1)
		strncpy(sppassword, tmpsppassword, sizeof(tmpsppassword));
	memset(tmpsppassword, 0, sizeof(tmpsppassword));
#endif

	return;
}
