static const char rcsid[] = "$Id: utils.c,v 1.11 2000/09/22 09:59:18 jurekb Exp $";
/*
 *  Solid POP3 - a POP3 server
 *  Copyright (C) 1999  Jerzy Balamut <jurekb@dione.ids.pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "includes.h"
#include "response.h"
#include "log.h"
#include "utils.h"
#include "configoptions.h"
#include <errno.h>
#ifdef HAVE_GRP_H
# include <grp.h>
#endif
#ifdef HASHSPOOL
# include "hashspool.h"
#endif

extern char username[];
extern char password[];
extern auth_method_t auth_method;
extern int tunnel[];
#ifdef MAPPING
extern char mapusername[];
extern int domapping;
#endif
#ifdef LOG_EXTEND
extern char ahname[];
#endif
#ifdef SPENCRYPTION
extern int encenabled;
# ifdef SPSASL
extern char sasl_mechanism[];
# endif
#endif
extern int pop_priority;
extern int pop_facility;

int expand_dir(char *dir, char *homedir, size_t maxlength) {
	/* maxlength <= PATH_MAX */
	char filename[PATH_MAX], *tmp, *tmp2;
	int breakwhile = 0;

	if (maxlength > sizeof(filename)) {
		pop_log(pop_priority, "utils: weird error - contact author "
			"of program!!!");
		return -1;
	}
	filename[0] = 0;
	tmp = dir;
	while(1) {
		if ((tmp2 = strchr(tmp, '%')) == NULL) {
			if ((strlen(filename) + strlen(tmp) + 1) > maxlength) {
				tmp2 = filename;
				break;
			}
			strcat(filename, tmp);
			break;
		}
		switch (tmp2[1]) {
			case '%':
				*(tmp2+1) = 0;
				if ((strlen(filename) + strlen(tmp) + 1) >
					maxlength) {
					breakwhile = 1;
					break;
				}
				strcat(filename, tmp);
				tmp = tmp2 + 2;
				break;
			case 's':
				
				*tmp2 = 0;
				if ((strlen(filename) + strlen(tmp) +
					strlen(username) + 1) > maxlength) {
				    breakwhile = 1;
				    break;
				}
				strcat(filename, tmp);
				strcat(filename, username);
				tmp = tmp2 + 2;
				break;
#ifdef MAPPING
			case 'm':
				*tmp2 = 0;
				if ((strlen(filename) + strlen(tmp) +
					strlen(mapusername) + 1) >
					maxlength) {
				    breakwhile = 1;
				    break;
				}
				strcat(filename, tmp);
				strcat(filename, mapusername);
				tmp = tmp2 + 2;
				break;
#endif
			case 'd':
				if ((tmp2[2] < '1') || (tmp2[2] > '8')) {
					breakwhile = 1;
					break;
				}
				*tmp2 = 0;
				if ((strlen(filename) + strlen(tmp) + 2) >
					maxlength) {
					breakwhile = 1;
					break;
				}
				strcat(filename, tmp);
				filename[(breakwhile = strlen(filename))] =
					username[tmp2[2] - '1'];
/* if username is too short, just ignore it - username is padded with NULs */
				filename[breakwhile + 1] = 0;
				breakwhile = 0;
				tmp = tmp2 + 3;
				break;
#ifdef HASHSPOOL
			case 'h':
				*tmp2 = 0;
				if (qpop_hash(username, filename, &tmp, 
					tmp2 + 2, maxlength) < 0)
					breakwhile = 1;
				break;
			case 'n':
				*tmp2 = 0;
				if (exim_nhash(username, filename, &tmp,
					tmp2 + 2, maxlength) < 0)
					breakwhile = 1;
				break;
			case 'p':
				*tmp2 = 0;
				if (pjw_hash(username, filename, &tmp,
					tmp2 + 2, maxlength, 1) < 0)
					breakwhile = 1;
				break;
			case 'P':
				*tmp2 = 0;
				if (pjw_hash(username, filename, &tmp,
					tmp2 + 2, maxlength, 0) < 0)
					breakwhile = 1;
				break;
#endif
			default:
				breakwhile = 1;
				break;
		}
		if (breakwhile)
			break;
	}
	if (tmp2 != NULL)
		return -1;
	dir[0] = 0;
	if ((filename[0] != '/') && (homedir != NULL)) {
		if ((strlen(homedir) + 2) > maxlength)
			return -1;
		strcpy(dir, homedir);
		strcat(dir, "/");
	}
	if ((strlen(dir) + strlen(filename) + 1) > maxlength)
		return -1;
	strcat(dir, filename);
	return 0;
}

ssize_t write_loop(int fd, void *abuf, size_t acount) {
	char *tmpbuf = (char *) abuf;
	size_t writed = 0;
	ssize_t tmp;
	
	if (acount == 0)
		return 0;
	while ((tmp = write(fd, tmpbuf + writed, acount - writed)) > 0) {
		writed += tmp;
		if (writed == acount)
			return writed;
	}
	if (tmp < 0) {
		pop_error("write_loop: write");
		return -1;
	}
	return writed;
}

int write_string(int fd, char *abuf) {
	size_t tmp = strlen(abuf);
	
	if (write_loop(fd, &tmp, sizeof(tmp)) != sizeof(tmp))
		return -1;
	if (write_loop(fd, abuf, tmp) != tmp)
		return -1;
	return 0;
}
ssize_t read_loop(int fd, void *abuf, size_t acount) {
/* This function works with root privileges */
	char *tmpbuf = (char *) abuf;
	ssize_t tmp, cread = 0;
	
	if (acount == 0)
		return 0;
	while ((tmp = read(fd, tmpbuf + cread, acount - cread)) > 0) {
		cread += tmp;
		if (cread == acount)
			return cread;
	}
	if (tmp < 0) {
		pop_error("read_loop: read");
		return -1;
	}
	return cread;
}
int read_string(int fd, char *abuf, size_t acount) {
/* This function works with root privileges */
	size_t tmp2;
	
	if (read_loop(fd, &tmp2, sizeof(tmp2)) != sizeof(tmp2))
		return -1;		
	if ((tmp2 >= acount) || (tmp2 < 0)) /* tmp2 is an unsigned number,
	so there wasn't security hole, however "szczezonego Pan Bog szczeze" */
		return -1;
	if (read_loop(fd, abuf, tmp2) != tmp2)
		return -1;
	abuf[tmp2] = 0;
	if (strlen(abuf) != tmp2)
		return -1;
	return 0;
}

int set_privileges(uid_t auid, gid_t agid) {
/* This function works with root privileges */
#ifdef HAVE_SETGROUPS
	gid_t groups[2];
	
	groups[0] = groups[1] = agid;
	if (setgroups(1, groups) < 0) {
		pop_error("setgroups");
		return -1;
	}
#endif
	if (setgid(agid) < 0) {
		pop_error("setgid");
		return -1;
	}
	if (setuid(auid) < 0) {
		pop_error("setuid");
		return -1;
	}
	return 0; /* I see no reason why this function could fail */
}

void after_authentication(int success, char *add_username) {
	char *tmp;
	
	tmp = ((add_username == NULL) ? "" : add_username);
	if (success < 0) {
#ifdef MAPPING
		if (domapping)
# ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "authentication failed for mapped user %.40s%.40s "
			    "- %.384s", username, tmp, ahname);
# else
			pop_log(pop_priority,
				"authentication failed for mapped user "
				"%.40s%.40s", username, tmp);
# endif
		else
# ifdef LOG_EXTEND
			pop_log(pop_priority,
				"authentication failed for user %.40s%.40s - "
				"%.384s", username, tmp, ahname);
# else
			pop_log(pop_priority,
				"authentication failed for user %.40s%.40s",
				username, tmp);
# endif
#else /* MAPPING */
# ifdef LOG_EXTEND
		pop_log(pop_priority,
			"authentication failed for user %.40s%.40s - %.384s",
			username, tmp, ahname);
# else
		pop_log(pop_priority,
			"authentication failed for user %.40s%.40s",
			username, tmp);
# endif
#endif /* MAPPING */
		sleep(1);
		send_error("authentication failed");
		exit(1);
	}
#ifdef MAPPING
	if (domapping)
# ifdef LOG_EXTEND
		pop_log(pop_priority,
			"mapped user %.40s%.40s authenticated - %.384s",
			username, tmp, ahname);
# else
		pop_log(pop_priority,
			"mapped user %.40s%.40s authenticated", username, tmp);
# endif
	else
# ifdef LOG_EXTEND
		pop_log(pop_priority,
			"user %.40s%.40s authenticated - %.384s",
			username, tmp, ahname);
# else
		pop_log(pop_priority,
			"user %.40s%.40s authenticated", username, tmp);
# endif
#else /* MAPPING */
# ifdef LOG_EXTEND
	pop_log(pop_priority,
		"user %.40s%.40s authenticated - %.384s", username, tmp,
		ahname);
# else
	pop_log(pop_priority,
		"user %.40s%.40s authenticated", username, tmp);
# endif

#endif /* MAPPING */
}

#ifdef APOP
void after_apop(int success, char *add_username) {
	char *tmp;

	tmp = ((add_username == NULL) ? "" : add_username);
	if (success < 0) {
# ifdef MAPPING
#  ifdef LOG_EXTEND
		if (domapping) {
			pop_log(pop_priority,
			    "APOP authentication failed for mapped user "
			    "%.40s.%.40s - %.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority,
			    "APOP authentication failed for mapped user "
			    "%.40s%.40s", username, tmp);
#  endif
		} else {
#  ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "APOP authentication failed for user "
			    "%.40s%.40s - %.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority,
		    	    "APOP authentication failed for user %.40s%.40s",
			    username, tmp);
#  endif
		}
# else /* MAPPING */
#  ifdef LOG_EXTEND
		pop_log(pop_priority,
			"APOP authentication failed for user %.40s%.40s - "
			"%.384s", username, tmp, ahname);
#  else
		pop_log(pop_priority,
			"APOP authentication failed for user %.40s%.40s",
			username, tmp);
#  endif
# endif /* MAPPING */
		sleep(1);
		send_error("authentication failed");
		exit(1);
	} else {
# ifdef MAPPING
		if (domapping) {
#  ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "mapped user %.40s%.40s authenticated through "
			    "APOP - %.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority,
			    "mapped user %.40s%.40s authenticated through APOP",
			    tmp, username);
#  endif

		} else {
#  ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "user %.40s%.40s authenticated through APOP - "
			    "%.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority,
			    "user %.40s%.40s authenticated through APOP",
			    username, tmp);
#  endif

		}
# else /* MAPPING */
#  ifdef LOG_EXTEND
		pop_log(pop_priority,
			"user %.40s%.40s authenticated through APOP - %.384s",
			username, tmp, ahname);
#  else
		pop_log(pop_priority,
			"user %.40s%.40s authenticated through APOP",
			username, tmp);
#  endif
# endif /* MAPPING */
	}
	
}
#endif /* APOP */

#ifdef SPSASL
void after_sasl(int success, char *add_username) {
	char *tmp;

	tmp = ((add_username == NULL) ? "" : add_username);
	if (success < 0) {
# ifdef MAPPING
		if (domapping) {
#  ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "SASL authentication failed for mapped user "
			    "%.40s.%.40s - %.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority,
			    "SASL authentication failed for mapped user "
			    "%.40s%.40s", username, tmp);
#  endif
		} else {
#  ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "SASL authentication failed for user %.40s%.40s "
			    "- %.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority, "SASL authentication failed for "
				"user %.40s%.40s", username, tmp);
#  endif
		}
# else /* MAPPING */
#  ifdef LOG_EXTEND
		pop_log(pop_priority,
			"SASL authentication failed for user %.40s%.40s - "
			"%.384s", username, tmp, ahname);
#  else
		pop_log(pop_priority,
			"SASL authentication failed for user %.40s%.40s",
			username, tmp);
#  endif
# endif /* MAPPING */
		sleep(1);
		send_error("authentication failed");
		exit(1);
	} else {
# ifdef MAPPING
		if (domapping) {
#  ifdef LOG_EXTEND
			pop_log(pop_priority, "mapped user %.40s%.40s "
			    "authenticated through SASL - %.384s", username,
			    tmp, ahname);
#  else
			pop_log(pop_priority,
			    "mapped user %.40s%.40s authenticated through "
			    "SASL", tmp, username);
#  endif

		} else {
#  ifdef LOG_EXTEND
			pop_log(pop_priority,
			    "user %.40s%.40s authenticated through SASL - "
			    "%.384s", username, tmp, ahname);
#  else
			pop_log(pop_priority,
			    "user %.40s%.40s authenticated through SASL",
			    username, tmp);
#  endif

		}
# else /* MAPPING */
#  ifdef LOG_EXTEND
		pop_log(pop_priority,
			"user %.40s%.40s authenticated through SASL - %.384s",
			username, tmp, ahname);
#  else
		pop_log(pop_priority,
			"user %.40s%.40s authenticated through SASL",
			username, tmp);
#  endif
# endif /* MAPPING */
	}
}
#endif /* SPSASL */

void write_auth_error(void) {
        pop_log(pop_priority, "can't write to pipe");
	pop_error("write");
	send_error("fatal error");
	exit(1);
}

#ifdef MAPPING
void write_general_auth(auth_method_t method, char *ausername, char *apassword,
	char *amapname) {
#else
void write_general_auth(auth_method_t method, char *ausername,
	char *apassword) {
#endif
	if (write_loop(tunnel[1], &method, sizeof(method)) < 0)
		write_auth_error();
	if (write_string(tunnel[1], ausername) < 0)
		write_auth_error();
#ifdef MAPPING
	if (write_string(tunnel[1], amapname) < 0)
		write_auth_error();
#endif
	if (write_string(tunnel[1], apassword) < 0)
		write_auth_error();
#ifdef SPENCRYPTION
	if (write_loop(tunnel[1], &encenabled, sizeof(encenabled)) < 0)
		write_auth_error();
# ifdef SPSASL
    /* it's required for CAPA mechanism, not only for SASL authentication */
	if (write_string(tunnel[1], sasl_mechanism) < 0)
		write_auth_error();
# endif
#endif
	write_configuration(tunnel[1]);
}

void read_auth_error(void) {
	pop_log(pop_priority, "child process corrupted");
	if (errno != 0)
		pop_error("read");
	exit(1);
}

void read_general_auth(void) {
	if (read_loop(tunnel[0], &auth_method, sizeof(auth_method))
		!= sizeof(auth_method))
		exit(1); /* probably QUIT in authentication state */
	if (read_string(tunnel[0], username, MAXARGLN + 1) < 0)
		read_auth_error();
#ifdef MAPPING
	if (read_string(tunnel[0], mapusername, MAXARGLN + 1) < 0)
		read_auth_error();
#endif
	if (read_string(tunnel[0], password, MAXARGLN + 1) < 0)
		read_auth_error();
#ifdef SPENCRYPTION
	if (read_loop(tunnel[0], &encenabled, sizeof(encenabled))
		!= sizeof(encenabled))
		read_auth_error();
# ifdef SPSASL
    /* it's required for CAPA mechanism, not only for SASL authentication */
	if (read_string(tunnel[0], sasl_mechanism, 256) < 0)
		read_auth_error();
# endif
#endif
	read_configuration(tunnel[0]);
}

struct str_logpriority {
	int facility;
	int priority;
};

void write_logpriority(void) {
	struct str_logpriority pr = {pop_facility, pop_priority};

	if (write_loop(tunnel[1], &pr, sizeof(pr)) < 0) {
	        pop_log(pop_priority, "can't write to pipe");
		pop_error("write");
		exit(1);
	}
}

void read_logpriority(void) {
	struct str_logpriority pr;

	if (read_loop(tunnel[0], &pr, sizeof(pr)) != sizeof(pr))
		read_auth_error();
	pop_priority = pr.priority;
	if (pop_facility != pr.facility) {
		pop_facility = pr.facility;
		pop_closelog();
		pop_openlog();
	}
}
